from .utils import TestCase
from .utils import remove_ansi
from .utils import transpile_header
from .utils import transpile_source


class Test(TestCase):

    def test_invalid_decorator_value(self):
        self.assert_transpile_raises(
            '@raises(A(B))\n'
            'def foo():\n'
            '    pass\n',
            '  File "", line 1\n'
            '    @raises(A(B))\n'
            '            ^\n'
            "CompileError: invalid decorator value\n")

    def test_invalid_decorator_value_syntax(self):
        self.assert_transpile_raises(
            '@raises[A]\n'
            'def foo():\n'
            '    pass\n',
            '  File "", line 1\n'
            '    @raises[A]\n'
            '     ^\n'
            "CompileError: decorators must be @name or @name()\n")

    def test_inferred_type_combined_integers_assignment_too_big(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    value = (0xffffffffffffffff + 1)\n'
            '    print(value)\n',
            '  File "", line 2\n'
            '        value = (0xffffffffffffffff + 1)\n'
            '                 ^\n'
            "CompileError: integer literal out of range for 'i64'\n")

    def test_test_functions_not_in_header(self):
        header = transpile_header('@test\n'
                                  'def test_foo():\n'
                                  '    pass\n',
                                  module_hpp='foo/lib.mys.hpp')

        self.assertNotIn('test_foo', header)

    def test_return_i64_from_function_returning_string(self):
        self.assert_transpile_raises(
            'def foo() -> string:\n'
            '    return True',
            '  File "", line 2\n'
            '        return True\n'
            '               ^\n'
            "CompileError: expected a 'string', got a 'bool'\n")

    def test_return_list_from_function_returning_tuple(self):
        self.assert_transpile_raises(
            'def foo() -> (bool, i64):\n'
            '    return [1]',
            '  File "", line 2\n'
            '        return [1]\n'
            '               ^\n'
            "CompileError: cannot convert list to '(bool, i64)'\n")

    def test_return_tuple_from_function_returning_list(self):
        self.assert_transpile_raises(
            'def foo() -> [bool]:\n'
            '    return (1, True)',
            '  File "", line 2\n'
            '        return (1, True)\n'
            '               ^\n'
            "CompileError: cannot convert tuple to '[bool]'\n")

    def test_wrong_number_of_function_parameters(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    pass\n'
            'def bar():\n'
            '    foo(1)\n',
            '  File "", line 4\n'
            '        foo(1)\n'
            '        ^\n'
            "CompileError: expected 0 parameters, got 1\n")

    def test_wrong_function_parameter_type(self):
        self.assert_transpile_raises(
            'def foo(a: string):\n'
            '    pass\n'
            'def bar():\n'
            '    foo(True)\n',
            '  File "", line 4\n'
            '        foo(True)\n'
            '            ^\n'
            "CompileError: expected a 'string', got a 'bool'\n")

    def test_compare_i64_and_bool(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return 1 == True',
            '  File "", line 2\n'
            '        return 1 == True\n'
            '               ^\n'
            "CompileError: cannot convert 'i64/i32/i16/i8/u64/u32/u16/u8' "
            "to 'bool'\n")

    def test_compare_mix_of_literals_and_known_types_1(self):
        source = transpile_source('def foo():\n'
                                  '    k: u64 = 1\n'
                                  '    v: i64 = 1\n'
                                  '    if 0xffffffffffffffff == k:\n'
                                  '        pass\n'
                                  '    print(v)\n')

        self.assert_in('18446744073709551615ull', source)

    def test_assign_256_to_u8(self):
        self.assert_transpile_raises(
            'A: u8 = 256\n',
            '  File "", line 1\n'
            '    A: u8 = 256\n'
            '            ^\n'
            "CompileError: integer literal out of range for 'u8'\n")

    def test_assign_over_max_to_u64(self):
        self.assert_transpile_raises(
            'A: u64 = 0x1ffffffffffffffff\n',
            '  File "", line 1\n'
            '    A: u64 = 0x1ffffffffffffffff\n'
            '             ^\n'
            "CompileError: integer literal out of range for 'u64'\n")

    def test_assign_max_to_i64(self):
        source = transpile_source('A: i64 = 0x7fffffffffffffff\n')

        self.assert_in('A = 9223372036854775807;', source)

    def test_assign_over_max_to_i64(self):
        self.assert_transpile_raises(
            'A: i64 = 0xffffffffffffffff\n',
            '  File "", line 1\n'
            '    A: i64 = 0xffffffffffffffff\n'
            '             ^\n'
            "CompileError: integer literal out of range for 'i64'\n")

    def test_assign_float_to_u8(self):
        self.assert_transpile_raises(
            'A: u8 = 2.0\n',
            '  File "", line 1\n'
            '    A: u8 = 2.0\n'
            '            ^\n'
            "CompileError: cannot convert float to 'u8'\n")

    def test_assign_negative_number_to_u32(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    k: u32 = -1\n'
            '    print(k)\n',
            '  File "", line 2\n'
            '        k: u32 = -1\n'
            '                 ^\n'
            "CompileError: integer literal out of range for 'u32'\n")

    def test_assign_positive_number_to_u32(self):
        source = transpile_source('def foo():\n'
                                  '    i: u32 = +1\n'
                                  '    j: u32 = --1\n'
                                  '    print(i, j)\n')

        self.assert_in('u32 i = 1;', source)
        self.assert_in('u32 j = 1;', source)

    def test_reassign_negative_number_to_u32(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    i: u32 = 0\n'
            '    i = -1\n'
            '    print(i)\n',
            '  File "", line 3\n'
            '        i = -1\n'
            '            ^\n'
            "CompileError: integer literal out of range for 'u32'\n")

    def test_arithmetics_on_mix_of_literals_and_known_types_too_big(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    k: i64 = 1\n'
            '    value = (0xffffffffffffffff + k)\n'
            '    print(value)\n',
            '  File "", line 3\n'
            '        value = (0xffffffffffffffff + k)\n'
            '                 ^\n'
            "CompileError: integer literal out of range for 'i64'\n")

    def test_arithmetics_on_mix_of_literals_and_known_types_negative(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    value: u8 = (-1 * 5)\n'
            '    print(value)\n',
            '  File "", line 2\n'
            '        value: u8 = (-1 * 5)\n'
            '                     ^\n'
            "CompileError: integer literal out of range for 'u8'\n")

    def test_change_integer_type(self):
        source = transpile_source('def foo():\n'
                                  '    value = (i8(-1) * i8(u32(5)))\n'
                                  '    print(value)\n')

        self.assert_in('value = (i8(-1) * i8(u32(5)));', source)

    def test_change_integer_type_error(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    value = (i8(-1) * u32(5))\n'
            '    print(value)\n',
            '  File "", line 2\n'
            '        value = (i8(-1) * u32(5))\n'
            '                 ^\n'
            "CompileError: types 'i8' and 'u32' differs\n")

    def test_function_call(self):
        source = transpile_source('def foo(a: i32, b: f32):\n'
                                  '    print(a, b)\n'
                                  'def bar():\n'
                                  '    foo(1, 2.1)\n')

        self.assert_in('foo(1, 2.1);', source)

    def test_tuple_unpack_variable_defined_other_type(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    def foo(self) -> (bool, i64):\n'
            '        return (True, -5)\n'
            'def foo():\n'
            '    v = Foo()\n'
            '    b: string = ""\n'
            '    a, b = v.foo()\n',
            '  File "", line 7\n'
            '        a, b = v.foo()\n'
            '           ^\n'
            "CompileError: expected a 'string', got a 'i64'\n")

    def test_tuple_unpack_integer(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a, b = 1\n'
            '    print(a, b)\n',
            '  File "", line 2\n'
            '        a, b = 1\n'
            '               ^\n'
            "CompileError: only tuples can be unpacked\n")

    def test_assert_between(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a = 2\n'
            '    assert 1 <= a < 3\n',
            '  File "", line 3\n'
            "        assert 1 <= a < 3\n"
            '               ^\n'
            "CompileError: can only compare two values\n")

    def test_compare_between(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a = 2\n'
            '    print(1 <= a < 3)\n',
            '  File "", line 3\n'
            "        print(1 <= a < 3)\n"
            '              ^\n'
            "CompileError: can only compare two values\n")

    def test_type_error_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: u16 = 1\n'
            '    b: u32 = 1\n'
            '    c = b - a\n'
            '    print(c)\n',
            '  File "", line 4\n'
            '        c = b - a\n'
            '            ^\n'
            "CompileError: types 'u32' and 'u16' differs\n")

    def test_type_error_2(self):
        self.assert_transpile_raises(
            'def foo(a: i64):\n'
            '    b: u64 = a\n'
            '    print(b)\n',
            '  File "", line 2\n'
            '        b: u64 = a\n'
            '                 ^\n'
            "CompileError: expected a 'u64', got a 'i64'\n")

    def test_type_error_3(self):
        self.assert_transpile_raises(
            'def bar() -> string:\n'
            '    return ""\n'
            'def foo():\n'
            '    b: u64 = bar()\n'
            '    print(b)\n',
            '  File "", line 4\n'
            '        b: u64 = bar()\n'
            '                 ^\n'
            "CompileError: expected a 'u64', got a 'string'\n")

    def test_type_error_4(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: u32 = 1\n'
            '    b = a - [1]\n'
            '    print(b)\n',
            '  File "", line 3\n'
            '        b = a - [1]\n'
            '            ^\n'
            "CompileError: cannot convert 'u32' to "
            "'[i64/i32/i16/i8/u64/u32/u16/u8]'\n")

    def test_type_error_5(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return\n',
            '  File "", line 2\n'
            '        return\n'
            '        ^\n'
            "CompileError: return value missing\n")

    def test_type_error_6(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    return 1\n',
            '  File "", line 2\n'
            '        return 1\n'
            '               ^\n'
            "CompileError: function does not return any value\n")

    def test_compare_wrong_types_1(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return 1 == [""]\n',
            '  File "", line 2\n'
            '        return 1 == [""]\n'
            '               ^\n'
            "CompileError: cannot convert 'i64/i32/i16/i8/u64/u32/u16/u8' to "
            "'[string]'\n")

    def test_compare_wrong_types_2(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return [""] in 1\n',
            '  File "", line 2\n'
            '        return [""] in 1\n'
            '                       ^\n'
            "CompileError: not an iterable\n")

    def test_compare_wrong_types_3(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return [""] not in 1\n',
            '  File "", line 2\n'
            '        return [""] not in 1\n'
            '                           ^\n'
            "CompileError: not an iterable\n")

    def test_compare_wrong_types_4(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return 2.0 == 1\n',
            '  File "", line 2\n'
            '        return 2.0 == 1\n'
            '               ^\n'
            "CompileError: cannot convert 'f64/f32' to "
            "'i64/i32/i16/i8/u64/u32/u16/u8'\n")

    def test_compare_wrong_types_5(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return 1.0 == [""]\n',
            '  File "", line 2\n'
            '        return 1.0 == [""]\n'
            '               ^\n'
            "CompileError: cannot convert 'f64/f32' to '[string]'\n")

    def test_compare_wrong_types_6(self):
        self.assert_transpile_raises(
            'def foo(a: i32) -> bool:\n'
            '    return a in [""]\n',
            '  File "", line 2\n'
            '        return a in [""]\n'
            '               ^\n'
            "CompileError: types 'i32' and 'string' differs\n")

    def test_compare_wrong_types_7(self):
        self.assert_transpile_raises(
            'def foo(a: i32) -> bool:\n'
            '    return a in a\n',
            '  File "", line 2\n'
            '        return a in a\n'
            '                    ^\n'
            "CompileError: not an iterable\n")

    def test_compare_wrong_types_8(self):
        self.assert_transpile_raises(
            'def foo(a: i32) -> bool:\n'
            '    return 1 in a\n',
            '  File "", line 2\n'
            '        return 1 in a\n'
            '                    ^\n'
            "CompileError: not an iterable\n")

    def test_compare_wrong_types_9(self):
        self.assert_transpile_raises(
            'def foo(a: i32) -> bool:\n'
            '    return "" == a\n',
            '  File "", line 2\n'
            '        return "" == a\n'
            '               ^\n'
            "CompileError: types 'string' and 'i32' differs\n")

    def test_compare_wrong_types_10(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(1 is None)\n',
            '  File "", line 2\n'
            '        print(1 is None)\n'
            '                   ^\n'
            "CompileError: 'i64' cannot be None\n")

    def test_compare_wrong_types_11(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(1.0 is None)\n',
            '  File "", line 2\n'
            '        print(1.0 is None)\n'
            '                     ^\n'
            "CompileError: 'f64' cannot be None\n")

    def test_compare_wrong_types_12(self):
        self.assert_transpile_raises(
            'def foo(a: i32):\n'
            '    print(a is None)\n',
            '  File "", line 2\n'
            '        print(a is None)\n'
            '              ^\n'
            "CompileError: 'i32' cannot be None\n")

    def test_compare_wrong_types_13(self):
        self.assert_transpile_raises(
            'def foo(a: i32):\n'
            '    print(None is a)\n',
            '  File "", line 2\n'
            '        print(None is a)\n'
            '              ^\n'
            "CompileError: 'i32' cannot be None\n")

    def test_compare_wrong_types_14(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(True is None)\n',
            '  File "", line 2\n'
            '        print(True is None)\n'
            '              ^\n'
            "CompileError: 'bool' cannot be None\n")

    def test_compare_wrong_types_15(self):
        self.assert_transpile_raises(
            'def foo(a: bool):\n'
            '    print(None is a)\n',
            '  File "", line 2\n'
            '        print(None is a)\n'
            '              ^\n'
            "CompileError: 'bool' cannot be None\n")

    def test_compare_wrong_types_16(self):
        self.assert_transpile_raises(
            'def foo(a: bool):\n'
            '    print(a is not 1)\n',
            '  File "", line 2\n'
            '        print(a is not 1)\n'
            '              ^\n'
            "CompileError: cannot convert 'bool' to "
            "'i64/i32/i16/i8/u64/u32/u16/u8'\n")

    def test_compare_wrong_types_17(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(None in [1, 5])\n',
            '  File "", line 2\n'
            '        print(None in [1, 5])\n'
            '              ^\n'
            "CompileError: 'i64' cannot be None\n")

    def test_compare_wrong_types_18(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(None == "")\n',
            '  File "", line 2\n'
            '        print(None == "")\n'
            '              ^\n'
            "CompileError: use 'is' and 'is not' to compare to None\n")

    def test_compare_wrong_types_20(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if (1, ("", True)) == (1, ("", 1)):\n'
            '        pass\n',
            # ToDo: Marker in wrong place.
            '  File "", line 2\n'
            '        if (1, ("", True)) == (1, ("", 1)):\n'
            '           ^\n'
            "CompileError: cannot convert 'bool' to "
            "'i64/i32/i16/i8/u64/u32/u16/u8'\n")

    def test_bool_op_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if True or None:\n'
            '        pass\n',
            '  File "", line 2\n'
            "        if True or None:\n"
            '                   ^\n'
            "CompileError: None is not a 'bool'\n")

    def test_bool_op_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    if True or False and 1:\n'
            '        pass\n',
            '  File "", line 2\n'
            "        if True or False and 1:\n"
            '                             ^\n'
            "CompileError: expected a 'bool', got a 'i64'\n")

    def test_inferred_type_none(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a = None\n',
            '  File "", line 2\n'
            "        a = None\n"
            '        ^\n'
            "CompileError: cannot infer type from None\n")

    def test_docstring(self):
        source = transpile_source('def foo():\n'
                                  '    "Hi!"\n')

        self.assert_in('void foo(void)\n'
                       '{\n'
                       '    __MYS_TRACEBACK_ENTER();\n'
                       '    __MYS_TRACEBACK_EXIT();\n'
                       '}\n',
                       source)

    def test_global_integer_out_of_range(self):
        self.assert_transpile_raises(
            'V: i8 = 1000\n',
            '  File "", line 1\n'
            "    V: i8 = 1000\n"
            '            ^\n'
            "CompileError: integer literal out of range for 'i8'\n")

    def test_global_wrong_value_type(self):
        self.assert_transpile_raises(
            'V: i8 = ""\n',
            '  File "", line 1\n'
            '    V: i8 = ""\n'
            '            ^\n'
            "CompileError: expected a 'i8', got a 'string'\n")

    def test_assign_to_wrong_type(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a = 1\n'
            '    a = ""\n',
            '  File "", line 3\n'
            '        a = ""\n'
            '            ^\n'
            "CompileError: expected a 'i64', got a 'string'\n")

    def test_min_wrong_types(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(min(u8(1), u8(2), i8(2)))\n',
            '  File "", line 2\n'
            '        print(min(u8(1), u8(2), i8(2)))\n'
            '                                ^\n'
            "CompileError: expected a 'u8', got a 'i8'\n")

    def test_min_no_args(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(min())\n',
            '  File "", line 2\n'
            '        print(min())\n'
            '              ^\n'
            "CompileError: expected at least one parameter\n")

    def test_min_wrong_return_type(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    assert min(1, 2) == i8(1)\n',
            '  File "", line 2\n'
            '        assert min(1, 2) == i8(1)\n'
            '               ^\n'
            "CompileError: types 'i64' and 'i8' differs\n")

    def test_max_wrong_types(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(max(i8(1), i8(2), u8(2)))\n',
            '  File "", line 2\n'
            '        print(max(i8(1), i8(2), u8(2)))\n'
            '                                ^\n'
            "CompileError: expected a 'i8', got a 'u8'\n")

    def test_max_no_args(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(max())\n',
            '  File "", line 2\n'
            '        print(max())\n'
            '              ^\n'
            "CompileError: expected at least one parameter\n")

    def test_max_wrong_return_type(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    assert max(u8(1), 2) == i8(1)\n',
            '  File "", line 2\n'
            '        assert max(u8(1), 2) == i8(1)\n'
            '               ^\n'
            "CompileError: types 'u8' and 'i8' differs\n")

    def test_len_no_params(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(len())\n',
            '  File "", line 2\n'
            '        print(len())\n'
            '              ^\n'
            "CompileError: expected 1 parameter, got 0\n")

    def test_len_two_params(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(len(1, 2))\n',
            '  File "", line 2\n'
            '        print(len(1, 2))\n'
            '              ^\n'
            "CompileError: expected 1 parameter, got 2\n")

    def test_len_compare_to_non_u64(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    assert len("") == i8(0)\n',
            '  File "", line 2\n'
            '        assert len("") == i8(0)\n'
            '               ^\n'
            "CompileError: types 'u64' and 'i8' differs\n")

    def test_str_compare_to_non_string(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    assert str(0) == i8(0)\n',
            '  File "", line 2\n'
            '        assert str(0) == i8(0)\n'
            '               ^\n'
            "CompileError: types 'string' and 'i8' differs\n")

    def test_tuple_index_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a = 1\n'
            '    v = (1, "b")\n'
            '    print(v[a])\n',
            '  File "", line 4\n'
            '        print(v[a])\n'
            '                ^\n'
            "CompileError: tuple indexes must be compile time known integers\n")

    def test_tuple_index_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    v = (1, "b")\n'
            '    print(v[1 / 2])\n',
            '  File "", line 3\n'
            '        print(v[1 / 2])\n'
            '                ^\n'
            "CompileError: tuple indexes must be compile time known integers\n")

    def test_tuple_item_assign_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    v = (1, "b")\n'
            '    v[0] = "ff"\n',
            '  File "", line 3\n'
            '        v[0] = "ff"\n'
            '               ^\n'
            "CompileError: expected a 'i64', got a 'string'\n")

    def test_tuple_item_assign_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a, b, c = (1, "b")\n',
            '  File "", line 2\n'
            '        a, b, c = (1, "b")\n'
            '        ^\n'
            "CompileError: expected 3 values to unpack, got 2\n")

    def test_return_nones_as_bool_in_tuple(self):
        self.assert_transpile_raises(
            'def foo() -> (bool, bool):\n'
            '    return (None, None)\n',
            '  File "", line 2\n'
            '        return (None, None)\n'
            '                ^\n'
            "CompileError: 'bool' cannot be None\n")

    def test_return_wrong_integer_in_tuple(self):
        self.assert_transpile_raises(
            'def foo() -> (bool, u8, string):\n'
            '    return (True, i8(1), "")\n',
            '  File "", line 2\n'
            '        return (True, i8(1), "")\n'
            '                      ^\n'
            "CompileError: expected a 'u8', got a 'i8'\n")

    def test_return_none_as_bool(self):
        self.assert_transpile_raises(
            'def foo() -> bool:\n'
            '    return None\n',
            '  File "", line 2\n'
            '        return None\n'
            '               ^\n'
            "CompileError: 'bool' cannot be None\n")

    def test_assign_none_to_i32(self):
        self.assert_transpile_raises(
            'A: i32 = None\n',
            '  File "", line 1\n'
            '    A: i32 = None\n'
            '             ^\n'
            "CompileError: 'i32' cannot be None\n")

    def test_return_short_tuple(self):
        self.assert_transpile_raises(
            'def foo() -> (bool, bool):\n'
            '    return (True, )\n',
            '  File "", line 2\n'
            '        return (True, )\n'
            '               ^\n'
            "CompileError: expected a '(bool, bool)', got a '(bool, )'\n")

    def test_tuple_index_out_of_range(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    v = ("", 1)\n'
            '    print(v[2])\n',
            '  File "", line 3\n'
            '        print(v[2])\n'
            '                ^\n'
            "CompileError: tuple index out of range\n")

    def test_assign_to_function_call(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def bar() -> i32:\n'
                             '    return 1\n'
                             'def foo():\n'
                             '    bar() = 1\n')

        self.assertEqual(
            remove_ansi(str(cm.exception)),
            '  File "<string>", line 4\n'
            '    bar() = 1\n'
            '    ^\n'
            "SyntaxError: cannot assign to function call\n")

    def test_call_iter_functions_outside_for(self):
        # At least true for now...
        for name in ['range', 'enumerate', 'zip', 'slice', 'reversed']:
            self.assert_transpile_raises(
                'def foo():\n'
                f'    v = {name}([1, 4])\n'
                '    print(v)',
                '  File "", line 2\n'
                f'        v = {name}([1, 4])\n'
                '            ^\n'
                "CompileError: function can only be used in for-loops\n")

    def test_main_in_non_main_file(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def main():\n'
                             '    pass\n')

        self.assertEqual(str(cm.exception), 'main() is only allowed in main.mys')

    def test_no_main_in_main_file(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo():\n'
                             '    pass\n',
                             has_main=True)

        self.assertEqual(str(cm.exception), 'main() not found in main.mys')

    def test_call_void_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    pass\n'
            'def bar():\n'
            '    foo().bar()\n',
            '  File "", line 4\n'
            '        foo().bar()\n'
            '        ^\n'
            "CompileError: None has no methods\n")

    def test_call_void_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    pass\n'
            'def bar():\n'
            '    print(foo())\n',
            '  File "", line 4\n'
            '        print(foo())\n'
            '              ^\n'
            "CompileError: None cannot be printed\n")

    def test_bare_compare(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    1 == 2\n',
            '  File "", line 2\n'
            '        1 == 2\n'
            '        ^\n'
            "CompileError: bare comparision\n")

    def test_bare_integer(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    1\n',
            '  File "", line 2\n'
            '        1\n'
            '        ^\n'
            "CompileError: bare integer\n")

    def test_bare_float(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    2.0\n',
            '  File "", line 2\n'
            '        2.0\n'
            '        ^\n'
            "CompileError: bare float\n")

    def test_bare_not(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    not True\n',
            '  File "", line 2\n'
            '        not True\n'
            '        ^\n'
            "CompileError: bare unary operation\n")

    def test_bare_add(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    1 + 2\n',
            '  File "", line 2\n'
            '        1 + 2\n'
            '        ^\n'
            "CompileError: bare binary operation\n")

    def test_bare_name(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    a: i32 = 0\n'
            '    a\n',
            '  File "", line 3\n'
            '        a\n'
            '        ^\n'
            "CompileError: bare name\n")

    def test_not_only_allowed_on_bool(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(not "hi")\n',
            '  File "", line 2\n'
            '        print(not "hi")\n'
            '                  ^\n'
            "CompileError: expected a 'bool', got a 'string'\n")

    def test_negative_bool(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(-True)\n',
            '  File "", line 2\n'
            '        print(-True)\n'
            '              ^\n'
            "CompileError: unary '-' can only operate on numbers\n")

    def test_named_parameter_wrong_name(self):
        self.assert_transpile_raises(
            'def foo(a: bool):\n'
            '    print(a)\n'
            'def bar():\n'
            '    foo(b=True)\n',
            '  File "", line 4\n'
            '        foo(b=True)\n'
            '            ^\n'
            "CompileError: invalid parameter 'b'\n")

    def test_named_parameter_twice(self):
        self.assert_transpile_raises(
            'def foo(a: bool, b: i64):\n'
            '    print(a, b)\n'
            'def bar():\n'
            '    foo(b=True, b=1)\n',
            '  File "", line 4\n'
            '        foo(b=True, b=1)\n'
            '                    ^\n'
            "CompileError: parameter 'b: i64' given more than once\n")

    def test_both_regular_and_named_parameter(self):
        self.assert_transpile_raises(
            'def foo(a: bool):\n'
            '    print(a)\n'
            'def bar():\n'
            '    foo(True, a=True)\n',
            '  File "", line 4\n'
            '        foo(True, a=True)\n'
            '                  ^\n'
            "CompileError: parameter 'a: bool' given more than once\n")

    def test_both_regular_and_named_parameter_2(self):
        self.assert_transpile_raises(
            'def foo(a: (bool, string)):\n'
            '    print(a)\n'
            'def bar():\n'
            '    foo(True, a=True)\n',
            '  File "", line 4\n'
            '        foo(True, a=True)\n'
            '                  ^\n'
            "CompileError: parameter 'a: (bool, string)' given more than once\n")

    def test_named_parameter_before_regular(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('def foo(a: bool, b: i64):\n'
                             '    print(a, b)\n'
                             'def bar():\n'
                             '    foo(a=True, 1)\n')

        self.assertEqual(
            remove_ansi(str(cm.exception)),
            '  File "<string>", line 4\n'
            '    foo(a=True, 1)\n'
            '                 ^\n'
            "SyntaxError: positional argument follows keyword argument\n")

    def test_name_clash(self):
        self.assert_transpile_raises(
            'def bar():\n'
            '    pass\n'
            'def foo():\n'
            '    bar = 1\n'
            '    print(bar)\n',
            '  File "", line 4\n'
            '        bar = 1\n'
            '        ^\n'
            "CompileError: redefining 'bar'\n")

    def test_inline_constant_default_bool_parameter_value(self):
        source = transpile_source('def foo(a: bool = True):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(mys::Bool(true))', source)

    def test_inline_constant_default_u8_parameter_value(self):
        source = transpile_source('def foo(a: u8 = 1):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(1)', source)

    def test_inline_constant_default_i8_parameter_value(self):
        source = transpile_source('def foo(a: i8 = -1):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(-1)', source)

    def test_inline_constant_default_f64_parameter_value(self):
        source = transpile_source('def foo(a: f64 = 5.1):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(5.1)', source)

    def test_inline_constant_default_string_parameter_value(self):
        source = transpile_source('def foo(a: string = "hi"):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(mys::String("hi"))', source)

    def test_inline_constant_default_class_parameter_value_none(self):
        source = transpile_source('class Foo:\n'
                                  '    pass\n'
                                  'def foo(a: Foo = None):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(nullptr)', source)

    def test_inline_constant_default_tuple_parameter_value_none(self):
        source = transpile_source('def foo(a: (i8, string) = None):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(nullptr)', source)

    def test_not_inline_constant_default_class_parameter_value_not_none(self):
        source = transpile_source('class Foo:\n'
                                  '    pass\n'
                                  'def foo(a: Foo = Foo()):\n'
                                  '    pass\n'
                                  'def bar():\n'
                                  '    foo()\n')

        self.assert_in('foo(mys::foo::lib::foo_a_default())', source)

    def test_aug_assign_wrong_integer_types(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    x: i32 = 0\n'
            '    x += i8(1)\n',
            '  File "", line 3\n'
            '        x += i8(1)\n'
            '             ^\n'
            "CompileError: expected a 'i32', got a 'i8'\n")

    def test_undefined_type_as_function_parameter(self):
        self.assert_transpile_raises(
            'def add(a: Foo):\n'
            '    pass\n',
            '  File "", line 1\n'
            '    def add(a: Foo):\n'
            '               ^\n'
            "CompileError: undefined type 'Foo'\n")

    def test_undefined_function_return_type(self):
        self.assert_transpile_raises(
            'def add() -> Foo:\n'
            '    return None\n',
            '  File "", line 1\n'
            '    def add() -> Foo:\n'
            '                 ^\n'
            "CompileError: undefined type 'Foo'\n")

    def test_power_f64_and_i64(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(2.0 ** 3)\n',
            '  File "", line 2\n'
            '        print(2.0 ** 3)\n'
            '              ^\n'
            "CompileError: cannot convert 'f64/f32' to "
            "'i64/i32/i16/i8/u64/u32/u16/u8'\n")

    # ToDo
    # def test_method_name_same_as_member_name(self):
    #     self.assert_transpile_raises(
    #         'class Bar:\n'
    #         '    text: string\n'
    #         '    def text(self) -> string:\n'
    #         '        return self.text\n',
    #         '  File "", line 1\n'
    #         '    def add() -> Foo:\n'
    #         '                 ^\n'
    #         "CompileError: undefined type 'Foo'\n")



from ..parser import ast
from .traits import is_trait_method_pure
from .utils import has_docstring


class CoverageTransformer(ast.NodeTransformer):
    """Traverses the AST and inserts coverage counters.

    """

    def __init__(self, source):
        self._variables = {}
        self._source_lines = source.splitlines()

    def variables(self):
        return sorted(self._variables.items())

    def visit_Module(self, node):
        body = [self.visit(item) for item in node.body]
        variables = []

        for _, variable in self.variables():
            variables.append(
                ast.AnnAssign(
                    target=ast.Name(id=variable, ctx=ast.Store()),
                    annotation=ast.Name(id='i64', ctx=ast.Load()),
                    value=ast.Constant(value=0, kind=None),
                    simple=1))

        node.body = variables + body

        return node

    def add_variable(self, lineno):
        variable = f'__MYS_COVERAGE_{lineno}'
        self._variables[lineno] = variable

        return ast.AugAssign(
            target=ast.Name(id=variable, ctx=ast.Store()),
            op=ast.Add(),
            value=ast.Constant(value=1, kind=None))

    def append_variable(self, body, node):
        self.append_variable_lineno(body, node.lineno)

    def append_variable_lineno(self, body, lineno):
        if lineno not in self._variables:
            body.append(self.add_variable(lineno))

    def visit_body(self, node, body=None):
        if body is None:
            body = []

        for item in node:
            self.append_variable(body, item)
            body.append(self.visit(item))

        return body

    def visit_trait(self, node):
        body = []

        for item in node.body:
            if isinstance(item, ast.FunctionDef) and is_trait_method_pure(item):
                body.append(item)
            else:
                body.append(self.visit(item))

        node.body = body

        return node

    def visit_ClassDef(self, node):
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Name):
                if decorator.id == 'trait':
                    return self.visit_trait(node)

        node.body = [self.visit(item) for item in node.body]

        return node

    def visit_FunctionDef(self, node):
        # Ignore tests in coverage.
        for decorator in node.decorator_list:
            if isinstance(decorator, ast.Name):
                if decorator.id == 'test':
                    return node

        body = []
        self.append_variable(body, node)

        body_iter = iter(node.body)

        if has_docstring(node):
            next(body_iter)

        node.body = self.visit_body(body_iter, body)

        return node

    def visit_For(self, node):
        body = []
        self.append_variable(body, node)
        node.body = self.visit_body(node.body, body)

        return node

    def visit_While(self, node):
        body = []
        self.append_variable(body, node)
        node.body = self.visit_body(node.body, body)

        return node

    def visit_If(self, node):
        node.body = self.visit_body(node.body)

        if node.orelse:
            last_body_lineno = node.body[-1].end_lineno
            first_orelse_lineno = node.orelse[0].lineno
            body = []

            for lineno in range(last_body_lineno, first_orelse_lineno - 1):
                if self._source_lines[lineno].strip() == 'else:':
                    self.append_variable_lineno(body, lineno + 1)

            node.orelse = self.visit_body(node.orelse, body)

        return node

    def visit_Match(self, node):
        for case in node.cases:
            body = [self.add_variable(case.pattern.lineno)]
            case.body = self.visit_body(case.body, body)

        return node

    def visit_Try(self, node):
        node.body = self.visit_body(node.body)

        for handler in node.handlers:
            handler.body = self.visit_body(handler.body)

        node.orelse = self.visit_body(node.orelse)
        node.finalbody = self.visit_body(node.finalbody)

        return node



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_complex(self):
        build_and_test_module('complex')

    def test_j_instead_of_i(self):
        self.assert_transpile_raises(
            'def foo():\n'
            "    print(1 + 2i)\n",
            '  File "", line 2\n'
            '        print(1 + 2i)\n'
            "                  ^\n"
            'CompileError: complex numbers are not supported\n')

    def test_complex_type(self):
        self.assert_transpile_raises(
            'VAR: complex = 1 + 3i\n',
            '  File "", line 1\n'
            '    VAR: complex = 1 + 3i\n'
            '         ^\n'
            "CompileError: undefined type 'complex'\n")



def get_source(source_lines, lineno, col_offset, end_lineno, end_col_offset):
    source = []
    number_of_lines = end_lineno - lineno + 1
    new_lineno = lineno

    if number_of_lines == 1:
        line = source_lines[lineno - 1]
        line = line[col_offset:end_col_offset]

        if line:
            source.append(line)
    else:
        for i in range(number_of_lines):
            line = source_lines[lineno + i - 1]

            if i == 0:
                line = line[col_offset:]

                if not line:
                    new_lineno += 1
                    continue
            elif i == number_of_lines - 1:
                line = line[:end_col_offset]

                if not line:
                    continue

            source.append(line)

    return new_lineno, source


def get_function_or_class_node_start(node):
    if node.decorator_list:
        node = node.decorator_list[0]

    return node.lineno, 0



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Python source expertise for coverage.py"""

import os.path
import types

from . import files
from .misc import CoverageException
from .misc import NoSource
from .phystokens import source_encoding
from .phystokens import source_token_lines
from .plugin import FileReporter


def read_python_source(filename):
    """Read the Python source text from `filename`.

    Returns bytes.

    """
    with open(filename, "rb") as fin:
        source = fin.read()

    return source.replace(b"\r\n", b"\n").replace(b"\r", b"\n")


def get_python_source(filename):
    """Return the source code, as unicode."""
    base, ext = os.path.splitext(filename)
    exts = [ext]

    for ext in exts:
        try_filename = base + ext
        if os.path.exists(try_filename):
            # A regular text file: open it.
            source = read_python_source(try_filename)
            break
    else:
        # Couldn't find source.
        exc_msg = "No source for code: '%s'.\n" % (filename,)
        exc_msg += "Aborting report output, consider using -i."
        raise NoSource(exc_msg)

    # Replace \f because of http://bugs.python.org/issue19035
    source = source.replace(b'\f', b' ')
    source = source.decode(source_encoding(source), "replace")

    # Python code should always end with a line with a newline.
    if source and source[-1] != '\n':
        source += '\n'

    return source


def source_for_file(filename):
    """Return the source filename for `filename`.

    Given a file name being traced, return the best guess as to the source
    file to attribute it to.

    """

    return filename


def source_for_morf(morf):
    """Get the source filename for the module-or-file `morf`."""
    if hasattr(morf, '__file__') and morf.__file__:
        filename = morf.__file__
    elif isinstance(morf, types.ModuleType):
        # A module should have had .__file__, otherwise we can't use it.
        # This could be a PEP-420 namespace package.
        raise CoverageException("Module {} has no file".format(morf))
    else:
        filename = morf

    filename = source_for_file(files.unicode_filename(filename))
    return filename


class PythonFileReporter(FileReporter):
    """Report support for a Python file."""

    def __init__(self, morf, coverage=None):
        self.coverage = coverage

        filename = source_for_morf(morf)

        super().__init__(files.canonical_filename(filename))

        if hasattr(morf, '__name__'):
            name = morf.__name__.replace(".", os.sep)
            if os.path.basename(filename).startswith('__init__.'):
                name += os.sep + "__init__"
            name += ".py"
            name = files.unicode_filename(name)
        else:
            name = files.relative_filename(filename)
        self.relname = name

        self._source = None
        self._parser = None
        self._excluded = None

    def relative_filename(self):
        return self.relname

    def source(self):
        if self._source is None:
            self._source = get_python_source(self.filename)
        return self._source

    def source_token_lines(self):
        return source_token_lines(self.source())



import os
import shutil
import subprocess
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import create_new_package
from .utils import remove_build_directory
from .utils import run_mys_command


class Test(TestCase):

    def run_safe_test_none(self, name):
        proc = subprocess.run(['./build/debug/test', name],
                              capture_output=True,
                              text=True)
        self.assertNotEqual(proc.returncode, 0)
        self.assert_in('Object is None.', proc.stderr)

    def run_unsafe_test_none(self, name):
        proc = subprocess.run(['./build/debug-unsafe/test', name],
                              capture_output=True,
                              text=True)
        self.assertNotEqual(proc.returncode, 0)
        self.assert_not_in('Object is None.', proc.stderr)

    def run_safe_test_index(self, name, message):
        proc = subprocess.run(['./build/debug/test', name],
                              capture_output=True,
                              text=True)
        self.assertNotEqual(proc.returncode, 0)
        self.assert_in(message, proc.stderr)

    def run_unsafe_test_index(self, name, message):
        proc = subprocess.run(['./build/debug-unsafe/test', name],
                              capture_output=True,
                              text=True)

        # The index tests in unsafe builds may or may not crash...
        if proc.returncode != 0:
            self.assert_not_in(message, proc.stderr)

    def test_safe(self):
        name = 'test_safe'
        remove_build_directory(name)

        shutil.copytree('tests/files/unsafe', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            try:
                with patch('sys.argv', ['mys', 'test', '-v']):
                    mys.cli.main()
            except SystemExit:
                pass

            self.run_safe_test_none('test_class_none_1')
            self.run_safe_test_none('test_class_none_2')
            self.run_safe_test_none('test_compare_lists_3')
            self.run_safe_test_none('test_add_lists')
            self.run_safe_test_none('test_list_acces_none')
            self.run_safe_test_none('test_list_len_of_none')
            self.run_safe_test_none('test_compare_tuples_3')
            self.run_safe_test_none('test_tuple_acces_none')
            self.run_safe_test_none('test_tuple_unpack_in_for_loop_none_element')
            self.run_safe_test_none('test_string_none')
            self.run_safe_test_none('test_string_len_of_none')
            self.run_safe_test_none('test_compare_dicts_3')
            self.run_safe_test_none('test_dict_acces_none')
            self.run_safe_test_none('test_dict_len_of_none')
            self.run_safe_test_none('test_set_none')
            self.run_safe_test_index('test_bytes_index',
                                     'Bytes index -99997 is out of range.')
            self.run_safe_test_index('test_list_pop_index',
                                     'Pop index 1000 is out of range.')
            self.run_safe_test_index('test_negative_list_index',
                                     'List index -99996 is out of range.')
            self.run_safe_test_index('test_string_get_char_at_index',
                                     'String index -99994 is out of range.')

    def test_unsafe(self):
        name = 'test_unsafe'
        remove_build_directory(name)

        shutil.copytree('tests/files/unsafe', f'tests/build/{name}')

        with Path(f'tests/build/{name}'):
            try:
                with patch('sys.argv', ['mys', 'test', '-v', '--unsafe']):
                    mys.cli.main()
            except SystemExit:
                pass

            self.run_unsafe_test_none('test_class_none_1')
            self.run_unsafe_test_none('test_compare_lists_3')
            self.run_unsafe_test_none('test_add_lists')
            self.run_unsafe_test_none('test_list_acces_none')
            self.run_unsafe_test_none('test_list_len_of_none')
            self.run_unsafe_test_none('test_compare_tuples_3')
            self.run_unsafe_test_none('test_tuple_acces_none')
            self.run_unsafe_test_none('test_tuple_unpack_in_for_loop_none_element')
            self.run_unsafe_test_none('test_string_none')
            self.run_unsafe_test_none('test_string_len_of_none')
            self.run_unsafe_test_none('test_compare_dicts_3')
            self.run_unsafe_test_none('test_dict_acces_none')
            self.run_unsafe_test_none('test_dict_len_of_none')
            self.run_unsafe_test_none('test_set_none')
            self.run_unsafe_test_index('test_bytes_index', 'bytes index out of range')
            self.run_unsafe_test_index('test_list_pop_index',
                                       'pop index out of range')
            self.run_unsafe_test_index('test_negative_list_index',
                                       'list index out of range')
            self.run_unsafe_test_index('test_string_get_char_at_index',
                                       'string index out of range')

    def test_unsafe_build_run_test(self):
        package_name = 'test_unsafe_build_run_test'
        remove_build_directory(package_name)
        create_new_package(package_name)
        path = os.getcwd()

        with Path(f'tests/build/{package_name}'):
            run_mys_command(['build', '-v', '--unsafe'], path)
            run_mys_command(['run', '-v', '--unsafe'], path)
            run_mys_command(['test', '-v', '--unsafe'], path)



"""
Sphinx Read the Docs Dark Mode
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Allows for toggable dark mode on the Read the Docs
theme for Sphinx.
"""

__title__ = "sphinx-rtd-dark-mode"
__description__ = "Dark mode for the Sphinx Read the Docs theme."
__author__ = "MrDogeBro"
__version__ = "1.2.3"
__license__ = "MIT"

from .dark_mode_loader import DarkModeLoader


def setup(app):
    app.add_config_value("default_dark_mode", True, "html")

    app.connect("config-inited", DarkModeLoader().configure)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_errors(self):
        build_and_test_module('errors')

    def test_bare_integer_in_try(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        a\n'
            '    except:\n'
            '        pass\n',
            '  File "", line 3\n'
            '            a\n'
            '            ^\n'
            "CompileError: bare name\n")

    def test_bare_integer_in_except(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        pass\n'
            '    except:\n'
            '        a\n',
            '  File "", line 5\n'
            '            a\n'
            '            ^\n'
            "CompileError: bare name\n")

    def test_try_except_different_variable_type_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        x = 1\n'
            '    except:\n'
            '        x = ""\n'
            '    print(x)\n',
            '  File "", line 6\n'
            '        print(x)\n'
            '              ^\n'
            "CompileError: undefined variable 'x'\n")

    def test_try_except_different_variable_type_2(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        x = 1\n'
            '    except GeneralError:\n'
            '        x = ""\n'
            '    except ValueError:\n'
            '        x = 2\n'
            '    print(x)\n',
            '  File "", line 8\n'
            '        print(x)\n'
            '              ^\n'
            "CompileError: undefined variable 'x'\n")

    def test_try_except_missing_branch(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        x = 1\n'
            '    except GeneralError:\n'
            '        pass\n'
            '    except ValueError:\n'
            '        x = 2\n'
            '    print(x)\n',
            '  File "", line 8\n'
            '        print(x)\n'
            '              ^\n'
            "CompileError: undefined variable 'x'\n")

    def test_all_branches_different_variable_type_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        if False:\n'
            '            x = 1\n'
            '        else:\n'
            '            x = 2\n'
            '    except GeneralError:\n'
            '        try:\n'
            '            x = 3\n'
            '        except:\n'
            '            if True:\n'
            '                x: u8 = 4\n'
            '            else:\n'
            '                x = 5\n'
            '    print(x)\n',
            '  File "", line 15\n'
            '        print(x)\n'
            '              ^\n'
            "CompileError: undefined variable 'x'\n")

    def test_missing_errors_in_raises(self):
        self.assert_transpile_raises(
            '@raises()\n'
            'def foo():\n'
            '    pass\n',
            '  File "", line 1\n'
            '    @raises()\n'
            '     ^\n'
            "CompileError: @raises requires at least one error\n")

    def test_raise_not_an_error_1(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    try:\n'
            '        raise 1\n'
            '    except:\n'
            '        pass\n',
            '  File "", line 3\n'
            '            raise 1\n'
            '                  ^\n'
            "CompileError: errors must implement the Error trait\n")

    def test_raise_not_an_error_2(self):
        self.assert_transpile_raises(
            'class Foo:\n'
            '    pass\n'
            'def foo():\n'
            '    try:\n'
            '        raise Foo()\n'
            '    except Foo:\n'
            '        pass\n',
            '  File "", line 5\n'
            '            raise Foo()\n'
            '                  ^\n'
            "CompileError: errors must implement the Error trait\n")



from ..parser import ast
from .comprehension import DictComprehension
from .comprehension import ListComprehension
from .comprehension import SetComprehension
from .constant_visitor import is_constant
from .generics import add_generic_class
from .utils import CompileError
from .utils import GenericType
from .utils import dot2ns
from .utils import format_mys_type
from .utils import is_float_literal
from .utils import is_integer_literal
from .utils import is_primitive_type
from .utils import make_float_literal
from .utils import make_integer_literal
from .utils import make_shared
from .utils import make_shared_dict
from .utils import make_shared_list
from .utils import make_shared_set
from .utils import mys_to_cpp_type
from .utils import raise_if_wrong_types
from .utils import raise_if_wrong_visited_type
from .utils import split_dict_mys_type


def is_allowed_dict_key_type(mys_type):
    if is_primitive_type(mys_type):
        return True
    elif mys_type == 'string':
        return True

    return False


class ValueCheckTypeVisitor:

    def __init__(self, visitor):
        self.visitor = visitor
        self.context = self.visitor.context

    def mys_to_cpp_type(self, mys_type):
        return mys_to_cpp_type(mys_type, self.context)

    def visit_tuple(self, node, mys_type):
        if not isinstance(mys_type, tuple):
            mys_type = format_mys_type(mys_type)

            raise CompileError(f"cannot convert tuple to '{mys_type}'", node)

        values = []
        types = []

        for i, item in enumerate(node.elts):
            values.append(self.visit(item, mys_type[i]))
            types.append(self.context.mys_type)

        raise_if_wrong_types(tuple(types), mys_type, node, self.context)
        self.context.mys_type = mys_type
        cpp_type = self.mys_to_cpp_type(mys_type)
        values = ", ".join(values)

        return make_shared(cpp_type[6:], values)

    def visit_list(self, node, mys_type):
        if not isinstance(mys_type, list):
            mys_type = format_mys_type(mys_type)

            raise CompileError(f"cannot convert list to '{mys_type}'", node)

        values = []
        item_mys_type = mys_type[0]

        for item in node.elts:
            values.append(self.visit(item, item_mys_type))

        self.context.mys_type = mys_type
        value = ", ".join(values)
        item_cpp_type = self.mys_to_cpp_type(item_mys_type)

        return make_shared_list(item_cpp_type, value)

    def visit_dict(self, node, mys_type):
        if isinstance(mys_type, set) and len(node.keys) == 0:
            item_mys_type = list(mys_type)[0]
            item_cpp_type = self.mys_to_cpp_type(item_mys_type)
            return make_shared_set(item_cpp_type, '')

        if not isinstance(mys_type, dict):
            mys_type = format_mys_type(mys_type)

            raise CompileError(f"cannot convert dict to '{mys_type}'", node)

        key_mys_type, value_mys_type = split_dict_mys_type(mys_type)

        if not is_allowed_dict_key_type(key_mys_type):
            raise CompileError("invalid key type", node)

        keys = []
        values = []

        for key, value in zip(node.keys, node.values):
            keys.append(self.visit(key, key_mys_type))
            values.append(self.visit(value, value_mys_type))

        self.context.mys_type = mys_type
        items = ", ".join([f'{{{key}, {value}}}' for key, value in zip(keys, values)])
        key_cpp_type = self.mys_to_cpp_type(key_mys_type)
        value_cpp_type = self.mys_to_cpp_type(value_mys_type)

        return make_shared_dict(key_cpp_type, value_cpp_type, items)

    def visit_set(self, node, mys_type):
        if not isinstance(mys_type, set):
            mys_type = format_mys_type(mys_type)

            raise CompileError(f"cannot convert set to '{mys_type}'", node)

        values = []
        item_mys_type = list(mys_type)[0]

        for item in node.elts:
            values.append(self.visit(item, item_mys_type))

        self.context.mys_type = mys_type
        value = ", ".join(values)
        item_cpp_type = self.mys_to_cpp_type(item_mys_type)

        return make_shared_set(item_cpp_type, value)

    def visit_list_comp(self, node, mys_type):
        return ListComprehension(node, mys_type, self.visitor).generate()

    def visit_dict_comp(self, node, mys_type):
        return DictComprehension(node, mys_type, self.visitor).generate()

    def visit_set_comp(self, node, mys_type):
        return SetComprehension(node, mys_type, self.visitor).generate()

    def visit_other(self, node, mys_type):
        value = self.visitor.visit(node)

        if self.context.is_trait_defined(mys_type):
            if self.context.is_class_defined(self.context.mys_type):
                definitions = self.context.get_class_definitions(
                    self.context.mys_type)

                if mys_type not in definitions.implements:
                    trait_type = format_mys_type(mys_type)
                    class_type = self.context.mys_type

                    raise CompileError(
                        f"'{class_type}' does not implement trait '{trait_type}'",
                        node)

                self.context.mys_type = mys_type
        elif self.context.is_trait_defined(self.context.mys_type):
            if self.context.is_class_defined(mys_type):
                value = f'static_cast<mys::shared_ptr<{dot2ns(mys_type)}>>({value})'
                definitions = self.context.get_class_definitions(mys_type)

                if self.context.mys_type not in definitions.implements:
                    trait_type = format_mys_type(self.context.mys_type)
                    class_type = mys_type

                    raise CompileError(
                        f"'{class_type}' does not implement trait '{trait_type}'",
                        node)
        else:
            if isinstance(mys_type, GenericType):
                mys_type = add_generic_class(mys_type.node, self.context)[1]

            raise_if_wrong_visited_type(self.context, mys_type, node)

        return value

    def visit(self, node, mys_type):
        if is_integer_literal(node):
            value = make_integer_literal(mys_type, node)
            self.context.mys_type = mys_type
        elif is_float_literal(node):
            value = make_float_literal(mys_type, node)
            self.context.mys_type = mys_type
        elif isinstance(node, ast.Tuple):
            value = self.visit_tuple(node, mys_type)
        elif isinstance(node, ast.List):
            value = self.visit_list(node, mys_type)
        elif isinstance(node, ast.Dict):
            value = self.visit_dict(node, mys_type)
        elif isinstance(node, ast.Set):
            value = self.visit_set(node, mys_type)
        elif isinstance(node, ast.ListComp):
            value = self.visit_list_comp(node, mys_type)
        elif isinstance(node, ast.DictComp):
            value = self.visit_dict_comp(node, mys_type)
        elif isinstance(node, ast.SetComp):
            value = self.visit_set_comp(node, mys_type)
        elif is_constant(node):
            value = self.visitor.visit(node)

            if isinstance(node.value, tuple) and len(node.value) == 1:
                pass
            else:
                if self.context.mys_type is None:
                    if not is_primitive_type(mys_type):
                        self.context.mys_type = mys_type

                raise_if_wrong_visited_type(self.context, mys_type, node)
        else:
            value = self.visit_other(node, mys_type)

        return value



from mys.transpiler import TranspilerError
from mys.transpiler.utils import make_function_name

from .utils import TestCase
from .utils import build_and_test_module
from .utils import remove_ansi
from .utils import transpile_source


class Test(TestCase):

    def test_functions(self):
        build_and_test_module('functions')

    def test_invalid_main_argument(self):
        with self.assertRaises(TranspilerError) as cm:
            transpile_source('def main(argv: i32): pass',
                             has_main=True)

        self.assert_exception_string(
            cm,
            '  File "", line 1\n'
            '    def main(argv: i32): pass\n'
            '    ^\n'
            "CompileError: main() takes 'argv: [string]' or no arguments\n")

    def test_main_no_argv(self):
        transpile_source('def main():\n'
                         '    pass\n',
                         has_main=True)

    def test_main_argv(self):
        transpile_source('def main(argv: [string]):\n'
                         '    pass\n',
                         has_main=True)

    def test_invalid_main_return_type(self):
        with self.assertRaises(TranspilerError) as cm:
            transpile_source('def main() -> i32: return 0',
                             has_main=True)

        self.assertEqual(remove_ansi(str(cm.exception)),
                         '  File "", line 1\n'
                         '    def main() -> i32: return 0\n'
                         '    ^\n'
                         "CompileError: main() must not return any value\n")

    def test_return_nothing_in_main(self):
        source = transpile_source('def main():\n'
                                  '    return\n',
                                  has_main=True)

        # main() must return void.
        self.assert_in('void main(int __argc, const char *__argv[])\n'
                       '{\n'
                       '    (void)__argc;\n'
                       '    (void)__argv;\n'
                       '    __MYS_TRACEBACK_ENTER();\n'
                       '    __MYS_TRACEBACK_SET(0);\n'
                       '    __MYS_TRACEBACK_EXIT();\n'
                       '    return;\n'
                       '    __MYS_TRACEBACK_EXIT();\n'
                       '}\n',
                       source)

    def test_lambda_not_supported(self):
        with self.assertRaises(TranspilerError) as cm:
            transpile_source('def main(): print((lambda x: x)(1))',
                             mys_path='foo.py',
                             has_main=True)

        self.assertEqual(remove_ansi(str(cm.exception)),
                         '  File "foo.py", line 1\n'
                         '    def main(): print((lambda x: x)(1))\n'
                         '                       ^\n'
                         'CompileError: lambda functions are not supported\n')

    def test_bad_syntax(self):
        with self.assertRaises(Exception) as cm:
            transpile_source('DEF main(): pass',
                             mys_path='<unknown>',
                             has_main=True)

        self.assertEqual(remove_ansi(str(cm.exception)),
                         '  File "<unknown>", line 1\n'
                         '    DEF main(): pass\n'
                         '        ^\n'
                         'SyntaxError: invalid syntax\n')

    def test_empty_function(self):
        source = transpile_source('def foo():\n'
                                  '    pass\n')

        self.assert_in('void foo(void)\n'
                       '{\n'
                       '    __MYS_TRACEBACK_ENTER();\n'
                       '    __MYS_TRACEBACK_SET(0);\n'
                       '\n'
                       '    __MYS_TRACEBACK_EXIT();\n'
                       '}\n',
                       source)

    def test_undefined_function(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    bar()\n',
            '  File "", line 2\n'
            '        bar()\n'
            '        ^\n'
            "CompileError: undefined function 'bar'\n")

    def test_test_can_not_take_any_values(self):
        self.assert_transpile_raises(
            '@test(H)\n'
            'def foo():\n'
            '    pass\n',
            '  File "", line 1\n'
            '    @test(H)\n'
            '     ^\n'
            "CompileError: no parameters expected\n")

    def test_non_snake_case_function(self):
        self.assert_transpile_raises(
            'def Apa():\n'
            '    pass\n',
            '  File "", line 1\n'
            '    def Apa():\n'
            '    ^\n'
            "CompileError: function names must be snake case\n")

    def test_non_snake_case_function_parameter_name(self):
        self.assert_transpile_raises(
            'def foo(A: i32):\n'
            '    pass\n',
            '  File "", line 1\n'
            '    def foo(A: i32):\n'
            '            ^\n'
            "CompileError: parameter names must be snake case\n")

    def test_missing_function_parameter_type(self):
        self.assert_transpile_raises(
            'def foo(x):\n'
            '    pass\n',
            '  File "", line 1\n'
            '    def foo(x):\n'
            '            ^\n'
            "CompileError: parameters must have a type\n")

    def test_test_function_with_parameter(self):
        self.assert_transpile_raises(
            '@test\n'
            'def test_foo(v: bool):\n'
            '    pass\n',
            '  File "", line 2\n'
            '    def test_foo(v: bool):\n'
            '    ^\n'
            "CompileError: test functions takes no parameters\n")

    def test_test_function_with_return_value(self):
        self.assert_transpile_raises(
            '@test\n'
            'def test_foo() -> bool:\n'
            '    return True\n',
            '  File "", line 2\n'
            '    def test_foo() -> bool:\n'
            '    ^\n'
            "CompileError: test functions must not return any value\n")

    def test_function_name(self):
        datas = [
            ('foo', [], None, 'foo'),
            ('foo', ['i64'], None, 'foo_i64'),
            ('foo', ['i64'], 'bool', 'foo_i64_r_bool'),
            ('foo',
             ['i64', ('bool', 'string')],
             ['bool'],
             'foo_i64_tb_bool_string_te_r_lb_bool_le'),
            ('kalle_kula',
             ['Foo', {'bool': 'Bar'}],
             None,
             'kalle_kula_Foo_db_bool_cn_Bar_de')
        ]

        for name, params, returns, expected in datas:
            actual = make_function_name(name, params, returns)
            self.assertEqual(actual, expected)

    def test_ambigious_call_in_assert(self):
        self.assert_transpile_raises(
            'def foo(a: i64) -> u8:\n'
            '    return 1\n'
            'def foo(a: i32) -> u8:\n'
            '    return 2\n'
            'def bar():\n'
            '    assert foo(1) == 1\n',
            '  File "", line 6\n'
            '        assert foo(1) == 1\n'
            '               ^\n'
            "CompileError: ambigious function call\n")

    def test_ambigious_call(self):
        self.assert_transpile_raises(
            'def foo(a: i64) -> u8:\n'
            '    return 1\n'
            'def foo(a: i32) -> u8:\n'
            '    return 2\n'
            'def bar():\n'
            '    foo(1)\n',
            '  File "", line 6\n'
            '        foo(1)\n'
            '        ^\n'
            "CompileError: ambigious function call\n")

    def test_ambigious_call_default(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    return 1\n'
            'def foo(a: i32 = 2) -> u8:\n'
            '    return 2\n'
            'def bar():\n'
            '    foo()\n',
            '  File "", line 6\n'
            '        foo()\n'
            '        ^\n'
            "CompileError: ambigious function call\n")

    def test_missing_return(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    pass\n',
            '  File "", line 1\n'
            '    def foo() -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")

    def test_missing_return_with_if(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    if True:\n'
            '        return 0\n'
            '    else:\n'
            '        pass',
            '  File "", line 1\n'
            '    def foo() -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")

    def test_missing_return_in_match_case(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    match 1:\n'
            '        case 1:\n'
            '            return 1\n'
            '        case 2:\n'
            '            pass\n',
            '  File "", line 1\n'
            '    def foo() -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")

    def test_missing_return_in_try(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    try:\n'
            '        pass\n'
            '    except:\n'
            '        return 0\n',
            '  File "", line 1\n'
            '    def foo() -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")

    def test_missing_return_in_while_true(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    while True:\n'
            '        break\n',
            '  File "", line 1\n'
            '    def foo() -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")

    def test_missing_return_in_while_true_if(self):
        self.assert_transpile_raises(
            'def foo() -> u8:\n'
            '    while True:\n'
            '        if False:\n'
            '            break\n',
            '  File "", line 1\n'
            '    def foo() -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")

    def test_missing_return_in_while_true_match(self):
        self.assert_transpile_raises(
            'def foo(v: u8) -> u8:\n'
            '    while True:\n'
            '        match v:\n'
            '            case 1:\n'
            '                pass\n'
            '            case 2:\n'
            '                pass\n',
            '  File "", line 1\n'
            '    def foo(v: u8) -> u8:\n'
            '    ^\n'
            "CompileError: missing return or raise\n")



import os
import shutil
import tarfile
from http.client import responses

import requests

from ..utils import Spinner
from ..utils import add_verbose_argument
from ..utils import read_package_configuration


def publish_create_release_package(name, version, archive):
    with Spinner(f'Creating {archive}'):
        base_dir = f'{name}-{version}'
        os.makedirs(base_dir)

        if os.path.exists('../../doc'):
            shutil.copytree('../../doc', f'{base_dir}/doc')

        shutil.copytree('../../src', f'{base_dir}/src')
        shutil.copy('../../package.toml', f'{base_dir}/package.toml')
        shutil.copy('../../README.rst', f'{base_dir}/README.rst')

        with tarfile.open(archive, 'w:gz') as fout:
            fout.add(base_dir)


def publish_upload_release_package(address, access_token, archive):
    with Spinner(f'Uploading {archive}'):
        with open(archive, 'rb') as fin:
            if access_token is None:
                params = None
            else:
                params = {'token': access_token}

            response = requests.post(f'{address}/package/{archive}',
                                     params=params,
                                     data=fin.read())

            if response.status_code != 200:
                raise Exception(
                    'Package publish failed with HTTP '
                    f'{response.status_code} {responses[response.status_code]}.')

    if response.text:
        response_json = response.json()

        if 'token' in response_json:
            print('Access token needed to publish this package again:',
                  response_json['token'])


def do_publish(_parser, args, _mys_config):
    config = read_package_configuration()
    publish_dir = 'build/publish'
    shutil.rmtree(publish_dir, ignore_errors=True)
    os.makedirs(publish_dir)
    path = os.getcwd()
    os.chdir(publish_dir)

    try:
        name = config['package']['name']
        version = config['package']['version']
        archive = f"{name}-{version}.tar.gz"
        publish_create_release_package(name, version, archive)
        publish_upload_release_package(args.address, args.token, archive)
    finally:
        os.chdir(path)


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'publish',
        description='Publish a release to the registry.')
    add_verbose_argument(subparser)
    subparser.add_argument('-a', '--address',
                           default='https://mys-lang.org',
                           help='Registry address (default: %(default)s)')
    subparser.add_argument('-t', '--token',
                           help='Package access token.')
    subparser.set_defaults(func=do_publish)



import getpass
import os
import re
import subprocess

from colors import cyan

from ..utils import BULB
from ..utils import ERROR
from ..utils import Spinner
from ..utils import box_print
from ..utils import create_file_from_template


class BadPackageNameError(Exception):
    pass


def git_config_get(item, default=None):
    try:
        return subprocess.check_output(['git', 'config', '--get', item],
                                       encoding='utf-8').strip()
    except Exception:
        return default


def create_new_file(path, **kwargs):
    create_file_from_template(path, 'new', **kwargs)


def validate_package_name(package_name):
    if not re.match(r'^[a-z][a-z0-9_]*$', package_name):
        raise BadPackageNameError()


def find_authors(authors):
    if authors is not None:
        return ', '.join([f'"{author}"' for author in authors])

    user = git_config_get('user.name', getpass.getuser())
    email = git_config_get('user.email', f'{user}@example.com')

    return f'"{user} <{email}>"'


def do_new(_parser, args, _mys_config):
    package_name = os.path.basename(args.path)
    package_name_title = package_name.replace('_', ' ').title()
    authors = find_authors(args.authors)

    try:
        with Spinner(text=f"Creating package {package_name}"):
            validate_package_name(package_name)

            os.makedirs(args.path)
            path = os.getcwd()
            os.chdir(args.path)

            try:
                create_new_file('package.toml',
                                package_name=package_name,
                                authors=authors)
                create_new_file('.gitignore')
                create_new_file('.gitattributes')
                create_new_file('README.rst',
                                package=package_name.replace('_', '-'),
                                title=package_name.replace('_', ' ').title(),
                                line='=' * len(package_name))
                create_new_file('LICENSE')
                os.mkdir('src')
                create_new_file('src/lib.mys')
                create_new_file('src/main.mys')
                os.makedirs('doc/modules')
                create_new_file('doc/index.rst',
                                package_name=package_name_title,
                                title_line='=' * len(package_name_title))
            finally:
                os.chdir(path)
    except BadPackageNameError:
        box_print(['Package names must start with a letter and only',
                   'contain letters, numbers and underscores. Only lower',
                   'case letters are allowed.',
                   '',
                   'Here are a few examples:',
                   '',
                   f'{cyan("mys new foo")}'
                   f'{cyan("mys new f1")}'
                   f'{cyan("mys new foo_bar")}'],
                  ERROR)
        raise Exception()

    cd = cyan(f'cd {package_name}')

    box_print(['Build and run the new package by typing:',
               '',
               f'{cd}',
               f'{cyan("mys run")}'],
              BULB,
              width=53)


def add_subparser(subparsers):
    subparser = subparsers.add_parser(
        'new',
        description='Create a new package.')
    subparser.add_argument(
        '--author',
        dest='authors',
        action='append',
        help=("Package author as 'Mys Lang <mys.lang@example.com>'. May "
              "be given multiple times."))
    subparser.add_argument('path')
    subparser.set_defaults(func=do_new)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Coverage data for coverage.py.

This file had the 4.x JSON data support, which is now gone.  This file still
has storage-agnostic helpers, and is kept to avoid changing too many imports.
CoverageData is now defined in sqldata.py, and imported here to keep the
imports working.

"""


def add_data_to_hash(data, filename, hasher):
    """Contribute `filename`'s data to the `hasher`.

    `hasher` is a `coverage.misc.Hasher` instance to be updated with
    the file's data.  It should only get the results data, not the run
    data.

    """
    hasher.update(sorted(data.lines(filename) or []))
    hasher.update(data.file_tracer(filename))



import argparse
import os
import sys
from traceback import print_exc

import toml
from colors import cyan

from ..version import __version__
from .subparsers import build
from .subparsers import clean
from .subparsers import delete
from .subparsers import doc
from .subparsers import fetch
from .subparsers import help as help_
from .subparsers import install
from .subparsers import new
from .subparsers import publish
from .subparsers import run
from .subparsers import style
from .subparsers import test
from .subparsers import transpile
from .utils import create_file

DESCRIPTION = f'''\
The Mys programming language package manager.

Available subcommands are:

    {cyan('new')}      Create a new package.
    {cyan('build')}    Build the appliaction.
    {cyan('run')}      Build and run the application.
    {cyan('test')}     Build and run tests
    {cyan('clean')}    Remove build output.
    {cyan('publish')}  Publish a release to the registry.
    {cyan('delete')}   Delete a package from the registry.
    {cyan('install')}  Install an application from local package or registry.
    {cyan('doc')}      Build the documentation.
    {cyan('fetch')}    Download and extract all dependencies.
    {cyan('style')}    Code styling.
'''


def find_config_file():
    path = os.getenv('MYS_CONFIG')
    config_dir = os.path.expanduser('~/.config/mys')
    config_path = os.path.join(config_dir, 'config.toml')

    if path is not None:
        return path

    if not os.path.exists(config_path):
        os.makedirs(config_dir, exist_ok=True)
        create_file(config_path, '')

    return config_path


def load_mys_config():
    """Mys tool configuration.

    Add validation when needed.

    """

    path = find_config_file()

    try:
        with open(path) as fin:
            return toml.loads(fin.read())
    except toml.decoder.TomlDecodeError:
        raise Exception(f"failed to load Mys configuration file '{path}'")


def create_parser():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument(
        '-C', '--directory',
        help='Change directory to given directory before doing anything.')
    parser.add_argument('--config', help='Configuration file to use.')
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    subparsers = parser.add_subparsers(dest='subcommand',
                                       help='Subcommand to execute.',
                                       metavar='subcommand')

    new.add_subparser(subparsers)
    build.add_subparser(subparsers)
    run.add_subparser(subparsers)
    test.add_subparser(subparsers)
    clean.add_subparser(subparsers)
    transpile.add_subparser(subparsers)
    publish.add_subparser(subparsers)
    delete.add_subparser(subparsers)
    install.add_subparser(subparsers)
    style.add_subparser(subparsers)
    doc.add_subparser(subparsers)
    fetch.add_subparser(subparsers)
    help_.add_subparser(subparsers)

    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()

    if not hasattr(args, 'func'):
        parser.print_help()
        sys.exit(1)

    if args.directory is not None:
        os.chdir(args.directory)

    try:
        args.func(parser, args, load_mys_config())
    except Exception as e:
        if args.debug:
            print_exc()

        sys.exit(str(e))
    except KeyboardInterrupt:
        print()

        if args.debug:
            raise

        sys.exit(1)



from .utils import TestCase
from .utils import build_and_test_module


class Test(TestCase):

    def test_string(self):
        build_and_test_module('string')

    def test_iterate_over_range_string(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i in range("a"):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i in range("a"):\n'
            '                       ^\n'
            "CompileError: parameter type must be an integer, not 'string'\n")

    def test_iterate_over_enumerate_string(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    for i, j in enumerate(range(2), ""):\n'
            '        print(i)\n',
            '  File "", line 2\n'
            '        for i, j in enumerate(range(2), ""):\n'
            '                                        ^\n'
            "CompileError: initial value must be an integer, not 'string'\n")

    def test_global_string(self):
        self.assert_transpile_raises(
            '"Hello!"\n',
            '  File "", line 1\n'
            '    "Hello!"\n'
            '    ^\n'
            "CompileError: syntax error\n")

    def test_string_member_access(self):
        self.assert_transpile_raises(
            'def foo(v: string):\n'
            '    v.a = 1\n',
            '  File "", line 2\n'
            '        v.a = 1\n'
            '        ^\n'
            "CompileError: 'string' has no member 'a'\n")

    def test_string_to_utf8_too_many_parameters(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    "".to_utf8(1)\n',
            '  File "", line 2\n'
            '        "".to_utf8(1)\n'
            "        ^\n"
            'CompileError: expected 0 parameters, got 1\n')

    def test_string_upper_too_many_parameters(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    "".upper(1)\n',
            '  File "", line 2\n'
            '        "".upper(1)\n'
            "        ^\n"
            'CompileError: expected 0 parameters, got 1\n')

    def test_string_lower_too_many_parameters(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    "".lower(1)\n',
            '  File "", line 2\n'
            '        "".lower(1)\n'
            "        ^\n"
            'CompileError: expected 0 parameters, got 1\n')

    def test_string_bad_method(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    "".foobar()\n',
            '  File "", line 2\n'
            '        "".foobar()\n'
            "        ^\n"
            'CompileError: string method not implemented\n')

    def test_positive_string(self):
        self.assert_transpile_raises(
            'def foo():\n'
            '    print(+"hi")\n',
            '  File "", line 2\n'
            '        print(+"hi")\n'
            '              ^\n'
            "CompileError: unary '+' can only operate on numbers\n")



import platform
import shutil
from unittest.mock import patch

import mys.cli

from .utils import Path
from .utils import TestCase
from .utils import build_and_test_module
from .utils import create_new_package
from .utils import remove_ansi
from .utils import remove_build_directory


class Test(TestCase):

    def test_calc(self):
        build_and_test_module('calc')

    def test_hello_world(self):
        build_and_test_module('hello_world')

    def test_various_1(self):
        build_and_test_module('various_1')

    def test_various_2(self):
        build_and_test_module('various_2')

    def test_various_3(self):
        build_and_test_module('various_3')

    def test_special_symbols(self):
        build_and_test_module('special_symbols')

    def test_filename_in_error_1(self):
        name = 'test_filename_in_error_1'
        remove_build_directory(name)
        shutil.copytree('tests/files/imports', f'tests/build/{name}')

        with Path(f'tests/build/{name}/mypkg'):
            with open("src/lib.mys", "a") as fout:
                fout.write("APA BANAN")

            with self.assertRaises(SystemExit) as cm:
                with patch('sys.argv', ['mys', 'build']):
                    mys.cli.main()

            self.assert_in('File "./src/lib.mys", line ',
                           remove_ansi(str(cm.exception)))

    def test_filename_in_error_2(self):
        name = 'test_filename_in_error_2'
        remove_build_directory(name)
        shutil.copytree('tests/files/imports', f'tests/build/{name}')

        with Path(f'tests/build/{name}/mypkg'):
            with open("src/lib.mys", "a") as fout:
                fout.write("A: bool = kaka()")

            with self.assertRaises(SystemExit) as cm:
                with patch('sys.argv', ['mys', 'build']):
                    mys.cli.main()

            self.assert_in('File "./src/lib.mys", line ',
                           remove_ansi(str(cm.exception)))

    def test_filename_in_error_in_local_dependency(self):
        name = 'test_filename_in_error_in_local_dependency'
        remove_build_directory(name)
        shutil.copytree('tests/files/imports', f'tests/build/{name}')

        with Path(f'tests/build/{name}/mypkg'):
            with open("../mypkg1/src/subpkg1/mod1.mys", "a") as fout:
                fout.write("APA BANAN")

            with self.assertRaises(SystemExit) as cm:
                with patch('sys.argv', ['mys', 'build']):
                    mys.cli.main()

            self.assert_in('File "../mypkg1/src/subpkg1/mod1.mys", line ',
                           remove_ansi(str(cm.exception)))

    def test_filename_in_error_in_dependency_from_registry(self):
        if platform.system() == 'Darwin':
            # Timestamps does not work well on fast Mac.
            return

        package_name = 'test_filename_in_error_in_dependency_from_registry'
        remove_build_directory(package_name)
        create_new_package(package_name)

        with Path(f'tests/build/{package_name}'):
            with open('package.toml', 'a') as fout:
                fout.write('bar = "0.3.0"\n')

            with patch('sys.argv', ['mys', 'build']):
                mys.cli.main()

            with open("build/dependencies/bar-0.3.0/src/lib.mys", "a") as fout:
                fout.write("APA BANAN")

            with self.assertRaises(SystemExit) as cm:
                with patch('sys.argv', ['mys', 'build']):
                    mys.cli.main()

            self.assert_in(
                'File "build/dependencies/bar-0.3.0/src/lib.mys", line ',
                remove_ansi(str(cm.exception)))



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Results of coverage measurement."""

from ..parser import ast
from ..transpiler.coverage_transformer import CoverageTransformer
from .debug import SimpleReprMixin


def find_statements(filename):
    with open(filename, 'r') as fin:
        source = fin.read()

    coverage_transformer = CoverageTransformer(source)
    coverage_transformer.visit(ast.parse(source))

    return {lineno for lineno, _ in coverage_transformer.variables()}


class Analysis:
    """The results of analyzing a FileReporter."""

    def __init__(self, data, file_reporter, file_mapper):
        self.data = data
        self.file_reporter = file_reporter
        self.filename = file_mapper(self.file_reporter.filename)
        self.statements = find_statements(self.filename)
        self.excluded = set()  # self.file_reporter.excluded_lines()

        # Identify missing statements.
        executed = set(self.data.lines(self.filename) or [])
        # executed = self.file_reporter.translate_lines(executed)
        self.executed = executed
        self.missing = self.statements - self.executed
        self._arc_possibilities = []
        self.exit_counts = {}
        self.no_branch = set()
        n_branches = n_partial_branches = n_missing_branches = 0

        self.numbers = Numbers(
            n_files=1,
            n_statements=len(self.statements),
            n_excluded=len(self.excluded),
            n_missing=len(self.missing),
            n_branches=n_branches,
            n_partial_branches=n_partial_branches,
            n_missing_branches=n_missing_branches,
        )


class Numbers(SimpleReprMixin):
    """The numerical results of measuring coverage.

    This holds the basic statistics from `Analysis`, and is used to roll
    up statistics across files.

    """
    # A global to determine the precision on coverage percentages, the number
    # of decimal places.
    _precision = 0
    _near0 = 1.0              # These will change when _precision is changed.
    _near100 = 99.0

    def __init__(self, n_files=0, n_statements=0, n_excluded=0, n_missing=0,
                    n_branches=0, n_partial_branches=0, n_missing_branches=0
                    ):
        self.n_files = n_files
        self.n_statements = n_statements
        self.n_excluded = n_excluded
        self.n_missing = n_missing
        self.n_branches = n_branches
        self.n_partial_branches = n_partial_branches
        self.n_missing_branches = n_missing_branches

    def init_args(self):
        """Return a list for __init__(*args) to recreate this object."""
        return [
            self.n_files, self.n_statements, self.n_excluded, self.n_missing,
            self.n_branches, self.n_partial_branches, self.n_missing_branches,
        ]

    @classmethod
    def set_precision(cls, precision):
        """Set the number of decimal places used to report percentages."""
        assert 0 <= precision < 10
        cls._precision = precision
        cls._near0 = 1.0 / 10**precision
        cls._near100 = 100.0 - cls._near0

    @property
    def n_executed(self):
        """Returns the number of executed statements."""
        return self.n_statements - self.n_missing

    @property
    def n_executed_branches(self):
        """Returns the number of executed branches."""
        return self.n_branches - self.n_missing_branches

    @property
    def pc_covered(self):
        """Returns a single percentage value for coverage."""
        if self.n_statements > 0:
            numerator, denominator = self.ratio_covered
            pc_cov = (100.0 * numerator) / denominator
        else:
            pc_cov = 100.0
        return pc_cov

    @property
    def pc_covered_str(self):
        """Returns the percent covered, as a string, without a percent sign.

        Note that "0" is only returned when the value is truly zero, and "100"
        is only returned when the value is truly 100.  Rounding can never
        result in either "0" or "100".

        """
        pc = self.pc_covered
        if 0 < pc < self._near0:
            pc = self._near0
        elif self._near100 < pc < 100:
            pc = self._near100
        else:
            pc = round(pc, self._precision)
        return "%.*f" % (self._precision, pc)

    @property
    def ratio_covered(self):
        """Return a numerator and denominator for the coverage ratio."""
        numerator = self.n_executed + self.n_executed_branches
        denominator = self.n_statements + self.n_branches
        return numerator, denominator

    def __add__(self, other):
        nums = Numbers()
        nums.n_files = self.n_files + other.n_files
        nums.n_statements = self.n_statements + other.n_statements
        nums.n_excluded = self.n_excluded + other.n_excluded
        nums.n_missing = self.n_missing + other.n_missing
        nums.n_branches = self.n_branches + other.n_branches
        nums.n_partial_branches = (
            self.n_partial_branches + other.n_partial_branches
            )
        nums.n_missing_branches = (
            self.n_missing_branches + other.n_missing_branches
            )
        return nums

    def __radd__(self, other):
        # Implementing 0+Numbers allows us to sum() a list of Numbers.
        if other == 0:
            return self
        return NotImplemented



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Sqlite coverage data."""

# TODO: factor out dataop debugging to a wrapper class?
# TODO: make sure all dataop debugging is in place somehow

import collections
import datetime
import glob
import itertools
import os
import re
import sqlite3
import sys
import zlib
from threading import get_ident as get_thread_id

from .debug import NoDebugging
from .debug import SimpleReprMixin
from .misc import CoverageException
from .misc import file_be_gone
from .misc import filename_suffix
from .numbits import numbits_to_nums
from .numbits import nums_to_numbits
from .version import __version__

# If you change the schema, increment the SCHEMA_VERSION, and update the
# docs in docs/dbschema.rst also.

SCHEMA_VERSION = 7

# Schema versions:
# 1: Released in 5.0a2
# 2: Added contexts in 5.0a3.
# 3: Replaced line table with line_map table.
# 4: Changed line_map.bitmap to line_map.numbits.
# 5: Added foreign key declarations.
# 6: Key-value in meta.
# 7: line_map -> line_bits

SCHEMA = """\
CREATE TABLE coverage_schema (
    -- One row, to record the version of the schema in this db.
    version integer
);

CREATE TABLE meta (
    -- Key-value pairs, to record metadata about the data
    key text,
    value text,
    unique (key)
    -- Keys:
    --  'has_arcs' boolean      -- Is this data recording branches?
    --  'sys_argv' text         -- The coverage command line that recorded the data.
    --  'version' text          -- The version of coverage.py that made the file.
    --  'when' text             -- Datetime when the file was created.
);

CREATE TABLE file (
    -- A row per file measured.
    id integer primary key,
    path text,
    unique (path)
);

CREATE TABLE context (
    -- A row per context measured.
    id integer primary key,
    context text,
    unique (context)
);

CREATE TABLE line_bits (
    -- If recording lines, a row per context per file executed.
    -- All of the line numbers for that file/context are in one numbits.
    file_id integer,            -- foreign key to `file`.
    context_id integer,         -- foreign key to `context`.
    numbits blob,               -- see the numbits functions in coverage.numbits
    foreign key (file_id) references file (id),
    foreign key (context_id) references context (id),
    unique (file_id, context_id)
);

CREATE TABLE arc (
    -- If recording branches, a row per context per from/to line transition executed.
    file_id integer,            -- foreign key to `file`.
    context_id integer,         -- foreign key to `context`.
    fromno integer,             -- line number jumped from.
    tono integer,               -- line number jumped to.
    foreign key (file_id) references file (id),
    foreign key (context_id) references context (id),
    unique (file_id, context_id, fromno, tono)
);

CREATE TABLE tracer (
    -- A row per file indicating the tracer used for that file.
    file_id integer primary key,
    tracer text,
    foreign key (file_id) references file (id)
);
"""

class CoverageData(SimpleReprMixin):
    """Manages collected coverage data, including file storage.

    This class is the public supported API to the data that coverage.py
    collects during program execution.  It includes information about what code
    was executed. It does not include information from the analysis phase, to
    determine what lines could have been executed, or what lines were not
    executed.

    .. note::

        The data file is currently a SQLite database file, with a
        :ref:`documented schema <dbschema>`. The schema is subject to change
        though, so be careful about querying it directly. Use this API if you
        can to isolate yourself from changes.

    There are a number of kinds of data that can be collected:

    * **lines**: the line numbers of source lines that were executed.
      These are always available.

    * **arcs**: pairs of source and destination line numbers for transitions
      between source lines.  These are only available if branch coverage was
      used.

    * **file tracer names**: the module names of the file tracer plugins that
      handled each file in the data.

    Lines, arcs, and file tracer names are stored for each source file. File
    names in this API are case-sensitive, even on platforms with
    case-insensitive file systems.

    A data file either stores lines, or arcs, but not both.

    A data file is associated with the data when the :class:`CoverageData`
    is created, using the parameters `basename`, `suffix`, and `no_disk`. The
    base name can be queried with :meth:`base_filename`, and the actual file
    name being used is available from :meth:`data_filename`.

    To read an existing coverage.py data file, use :meth:`read`.  You can then
    access the line, arc, or file tracer data with :meth:`lines`, :meth:`arcs`,
    or :meth:`file_tracer`.

    The :meth:`has_arcs` method indicates whether arc data is available.  You
    can get a set of the files in the data with :meth:`measured_files`.  As
    with most Python containers, you can determine if there is any data at all
    by using this object as a boolean value.

    The contexts for each line in a file can be read with
    :meth:`contexts_by_lineno`.

    To limit querying to certain contexts, use :meth:`set_query_context` or
    :meth:`set_query_contexts`. These will narrow the focus of subsequent
    :meth:`lines`, :meth:`arcs`, and :meth:`contexts_by_lineno` calls. The set
    of all measured context names can be retrieved with
    :meth:`measured_contexts`.

    Most data files will be created by coverage.py itself, but you can use
    methods here to create data files if you like.  The :meth:`add_lines`,
    :meth:`add_arcs`, and :meth:`add_file_tracers` methods add data, in ways
    that are convenient for coverage.py.

    To record data for contexts, use :meth:`set_context` to set a context to
    be used for subsequent :meth:`add_lines` and :meth:`add_arcs` calls.

    To add a source file without any measured data, use :meth:`touch_file`,
    or :meth:`touch_files` for a list of such files.

    Write the data to its file with :meth:`write`.

    You can clear the data in memory with :meth:`erase`.  Two data collections
    can be combined by using :meth:`update` on one :class:`CoverageData`,
    passing it the other.

    Data in a :class:`CoverageData` can be serialized and deserialized with
    :meth:`dumps` and :meth:`loads`.

    """

    def __init__(self,
                 basename=None,
                 suffix=None,
                 no_disk=False,
                 warn=None,
                 debug=None):
        """Create a :class:`CoverageData` object to hold coverage-measured data.

        Arguments:
            basename (str): the base name of the data file, defaulting to
                ".coverage".
            suffix (str or bool): has the same meaning as the `data_suffix`
                argument to :class:`coverage.Coverage`.
            no_disk (bool): if True, keep all data in memory, and don't
                write any disk file.
            warn: a warning callback function, accepting a warning message
                argument.
            debug: a `DebugControl` object (optional)

        """
        self._no_disk = no_disk
        self._basename = os.path.abspath(basename or ".coverage")
        self._suffix = suffix
        self._warn = warn
        self._debug = debug or NoDebugging()

        self._choose_filename()
        self._file_map = {}
        # Maps thread ids to SqliteDb objects.
        self._dbs = {}
        self._pid = os.getpid()

        # Are we in sync with the data file?
        self._have_used = False

        self._has_lines = False
        self._has_arcs = False

        self._current_context = None
        self._current_context_id = None
        self._query_context_ids = None

    def _choose_filename(self):
        """Set self._filename based on inited attributes."""
        if self._no_disk:
            self._filename = ":memory:"
        else:
            self._filename = self._basename
            suffix = filename_suffix(self._suffix)
            if suffix:
                self._filename += "." + suffix

    def _reset(self):
        """Reset our attributes."""
        if self._dbs:
            for db in self._dbs.values():
                db.close()
        self._dbs = {}
        self._file_map = {}
        self._have_used = False
        self._current_context_id = None

    def _create_db(self):
        """Create a db file that doesn't exist yet.

        Initializes the schema and certain metadata.
        """
        if self._debug.should('dataio'):
            self._debug.write("Creating data file {!r}".format(self._filename))
        self._dbs[get_thread_id()] = db = SqliteDb(self._filename, self._debug)
        with db:
            db.executescript(SCHEMA)
            db.execute("insert into coverage_schema (version) values (?)",
                       (SCHEMA_VERSION,))
            db.executemany(
                "insert into meta (key, value) values (?, ?)",
                [
                    ('sys_argv', str(getattr(sys, 'argv', None))),
                    ('version', __version__),
                    ('when', datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')),
                ]
            )

    def _open_db(self):
        """Open an existing db file, and read its metadata."""
        if self._debug.should('dataio'):
            self._debug.write("Opening data file {!r}".format(self._filename))
        self._dbs[get_thread_id()] = SqliteDb(self._filename, self._debug)
        self._read_db()

    def _read_db(self):
        """Read the metadata from a database so that we are ready to use it."""
        with self._dbs[get_thread_id()] as db:
            try:
                schema_version, = db.execute_one(
                    "select version from coverage_schema")
            except Exception as exc:
                raise CoverageException(
                    "Data file {!r} doesn't seem to be a coverage data "
                    "file: {}".format(self._filename, exc))
            else:
                if schema_version != SCHEMA_VERSION:
                    raise CoverageException(
                        "Couldn't use data file {!r}: wrong schema: {} instead "
                        "of {}".format(self._filename,
                                       schema_version,
                                       SCHEMA_VERSION))

            for row in db.execute("select value from meta where key = 'has_arcs'"):
                self._has_arcs = bool(int(row[0]))
                self._has_lines = not self._has_arcs

            for path, file_id in db.execute("select path, id from file"):
                self._file_map[path] = file_id

    def _connect(self):
        """Get the SqliteDb object to use."""
        if get_thread_id() not in self._dbs:
            if os.path.exists(self._filename):
                self._open_db()
            else:
                self._create_db()
        return self._dbs[get_thread_id()]

    def __nonzero__(self):
        if (get_thread_id() not in self._dbs and not os.path.exists(self._filename)):
            return False
        try:
            with self._connect() as con:
                rows = con.execute("select * from file limit 1")
                return bool(list(rows))
        except CoverageException:
            return False

    __bool__ = __nonzero__

    def dumps(self):
        """Serialize the current data to a byte string.

        The format of the serialized data is not documented. It is only
        suitable for use with :meth:`loads` in the same version of
        coverage.py.

        Returns:
            A byte string of serialized data.

        .. versionadded:: 5.0

        """
        if self._debug.should('dataio'):
            self._debug.write(
                "Dumping data from data file {!r}".format(
                    self._filename))
        with self._connect() as con:
            return b'z' + zlib.compress(con.dump())

    def loads(self, data):
        """Deserialize data from :meth:`dumps`

        Use with a newly-created empty :class:`CoverageData` object.  It's
        undefined what happens if the object already has data in it.

        Arguments:
            data: A byte string of serialized data produced by :meth:`dumps`.

        .. versionadded:: 5.0

        """
        if self._debug.should('dataio'):
            self._debug.write(
                "Loading data into data file {!r}".format(
                    self._filename))
        if data[:1] != b'z':
            raise CoverageException(
                "Unrecognized serialization: {!r} (head of {} bytes)".format(
                    data[:40],
                    len(data)))
        script = zlib.decompress(data[1:])
        self._dbs[get_thread_id()] = db = SqliteDb(self._filename, self._debug)
        with db:
            db.executescript(script)
        self._read_db()
        self._have_used = True

    def _file_id(self, filename, add=False):
        """Get the file id for `filename`.

        If filename is not in the database yet, add it if `add` is True.
        If `add` is not True, return None.
        """
        if filename not in self._file_map:
            if add:
                with self._connect() as con:
                    cur = con.execute(
                        "insert or replace into file (path) values (?)",
                        (filename,))
                    self._file_map[filename] = cur.lastrowid
        return self._file_map.get(filename)

    def _context_id(self, context):
        """Get the id for a context."""
        assert context is not None
        self._start_using()
        with self._connect() as con:
            row = con.execute_one("select id from context where context = ?",
                                  (context,))
            if row is not None:
                return row[0]
            else:
                return None

    def set_context(self, context):
        """Set the current context for future :meth:`add_lines` etc.

        `context` is a str, the name of the context to use for the next data
        additions.  The context persists until the next :meth:`set_context`.

        .. versionadded:: 5.0

        """
        if self._debug.should('dataop'):
            self._debug.write("Setting context: %r" % (context,))
        self._current_context = context
        self._current_context_id = None

    def _set_context_id(self):
        """Use the _current_context to set _current_context_id."""
        context = self._current_context or ""
        context_id = self._context_id(context)
        if context_id is not None:
            self._current_context_id = context_id
        else:
            with self._connect() as con:
                cur = con.execute("insert into context (context) values (?)",
                                  (context,))
                self._current_context_id = cur.lastrowid

    def base_filename(self):
        """The base filename for storing data.

        .. versionadded:: 5.0

        """
        return self._basename

    def data_filename(self):
        """Where is the data stored?

        .. versionadded:: 5.0

        """
        return self._filename

    def add_lines(self, line_data):
        """Add measured line data.

        `line_data` is a dictionary mapping file names to dictionaries::

            { filename: { lineno: None, ... }, ...}

        """
        if self._debug.should('dataop'):
            self._debug.write("Adding lines: %d files, %d lines total" % (
                len(line_data), sum(len(lines) for lines in line_data.values())
            ))
        self._start_using()
        self._choose_lines_or_arcs(lines=True)
        if not line_data:
            return
        with self._connect() as con:
            self._set_context_id()
            for filename, linenos in line_data.items():
                linemap = nums_to_numbits(linenos)
                file_id = self._file_id(filename, add=True)
                con.execute(
                    "insert or replace into line_bits "
                    " (file_id, context_id, numbits) values (?, ?, ?)",
                    (file_id, self._current_context_id, linemap),
                )

    def add_arcs(self, arc_data):
        """Add measured arc data.

        `arc_data` is a dictionary mapping file names to dictionaries::

            { filename: { (l1,l2): None, ... }, ...}

        """
        if self._debug.should('dataop'):
            self._debug.write("Adding arcs: %d files, %d arcs total" % (
                len(arc_data), sum(len(arcs) for arcs in arc_data.values())
            ))
        self._start_using()
        self._choose_lines_or_arcs(arcs=True)
        if not arc_data:
            return
        with self._connect() as con:
            self._set_context_id()
            for filename, arcs in arc_data.items():
                file_id = self._file_id(filename, add=True)
                data = [(file_id, self._current_context_id, fromno, tono)
                        for fromno, tono in arcs]
                con.executemany(
                    "insert or ignore into arc "
                    "(file_id, context_id, fromno, tono) values (?, ?, ?, ?)",
                    data,
                )

    def _choose_lines_or_arcs(self, lines=False, arcs=False):
        """Force the data file to choose between lines and arcs."""
        assert lines or arcs
        assert not (lines and arcs)
        if lines and self._has_arcs:
            raise CoverageException("Can't add lines to existing arc data")
        if arcs and self._has_lines:
            raise CoverageException("Can't add arcs to existing line data")
        if not self._has_arcs and not self._has_lines:
            self._has_lines = lines
            self._has_arcs = arcs
            with self._connect() as con:
                con.execute(
                    "insert into meta (key, value) values (?, ?)",
                    ('has_arcs', str(int(arcs)))
                )

    def add_file_tracers(self, file_tracers):
        """Add per-file plugin information.

        `file_tracers` is { filename: plugin_name, ... }

        """
        if self._debug.should('dataop'):
            self._debug.write("Adding file tracers: %d files" % (len(file_tracers),))
        if not file_tracers:
            return
        self._start_using()
        with self._connect() as con:
            for filename, plugin_name in file_tracers.items():
                file_id = self._file_id(filename)
                if file_id is None:
                    raise CoverageException(
                        "Can't add file tracer data for unmeasured file '%s'" % (
                            filename,))

                existing_plugin = self.file_tracer(filename)
                if existing_plugin:
                    if existing_plugin != plugin_name:
                        raise CoverageException(
                            "Conflicting file tracer name for '%s': %r vs %r" % (
                                filename, existing_plugin, plugin_name,
                            )
                        )
                elif plugin_name:
                    con.execute(
                        "insert into tracer (file_id, tracer) values (?, ?)",
                        (file_id, plugin_name)
                    )

    def touch_file(self, filename, plugin_name=""):
        """Ensure that `filename` appears in the data, empty if needed.

        `plugin_name` is the name of the plugin responsible for this file. It is used
        to associate the right filereporter, etc.
        """
        self.touch_files([filename], plugin_name)

    def touch_files(self, filenames, plugin_name=""):
        """Ensure that `filenames` appear in the data, empty if needed.

        `plugin_name` is the name of the plugin responsible for these
        files. It is used to associate the right filereporter, etc.

        """
        if self._debug.should('dataop'):
            self._debug.write("Touching %r" % (filenames,))
        self._start_using()
        with self._connect(): # Use this to get one transaction.
            if not self._has_arcs and not self._has_lines:
                raise CoverageException("Can't touch files in an empty CoverageData")

            for filename in filenames:
                self._file_id(filename, add=True)
                if plugin_name:
                    # Set the tracer for this file
                    self.add_file_tracers({filename: plugin_name})

    def update(self, other_data, aliases=None):
        """Update this data with data from several other :class:`CoverageData`
        instances.

        If `aliases` is provided, it's a `PathAliases` object that is used to
        re-map paths to match the local machine's.

        """
        if self._debug.should('dataop'):
            self._debug.write("Updating with data from %r" % (
                getattr(other_data, '_filename', '???'),
            ))
        if self._has_lines and other_data._has_arcs:
            raise CoverageException("Can't combine arc data with line data")
        if self._has_arcs and other_data._has_lines:
            raise CoverageException("Can't combine line data with arc data")

        # Force the database we're writing to to exist before we start nesting
        # contexts.
        self._start_using()

        # Collector for all arcs, lines and tracers
        other_data.read()
        with other_data._connect() as conn:
            # Get files data.
            cur = conn.execute('select path from file')
            files = {path: aliases.map(path) for (path,) in cur}
            cur.close()

            # Get contexts data.
            cur = conn.execute('select context from context')
            contexts = [context for (context,) in cur]
            cur.close()

            # Get arc data.
            cur = conn.execute(
                'select file.path, context.context, arc.fromno, arc.tono '
                'from arc '
                'inner join file on file.id = arc.file_id '
                'inner join context on context.id = arc.context_id'
            )
            arcs = [(files[path], context, fromno, tono)
                    for (path, context, fromno, tono) in cur]
            cur.close()

            # Get line data.
            cur = conn.execute(
                'select file.path, context.context, line_bits.numbits '
                'from line_bits '
                'inner join file on file.id = line_bits.file_id '
                'inner join context on context.id = line_bits.context_id'
                )
            lines = {
                (files[path], context): numbits
                for (path, context, numbits) in cur
                }
            cur.close()

            # Get tracer data.
            cur = conn.execute(
                'select file.path, tracer '
                'from tracer '
                'inner join file on file.id = tracer.file_id'
            )
            tracers = {files[path]: tracer for (path, tracer) in cur}
            cur.close()

        with self._connect() as conn:
            conn.con.isolation_level = 'IMMEDIATE'

            # Get all tracers in the DB. Files not in the tracers are assumed
            # to have an empty string tracer. Since Sqlite does not support
            # full outer joins, we have to make two queries to fill the
            # dictionary.
            this_tracers = {path: ''
                            for path, in conn.execute('select path from file')}
            this_tracers.update({
                aliases.map(path): tracer
                for path, tracer in conn.execute(
                    'select file.path, tracer from tracer '
                    'inner join file on file.id = tracer.file_id'
                )
            })

            # Create all file and context rows in the DB.
            conn.executemany(
                'insert or ignore into file (path) values (?)',
                ((file,) for file in files.values())
            )
            file_ids = {
                path: id
                for id, path in conn.execute('select id, path from file')
            }
            conn.executemany(
                'insert or ignore into context (context) values (?)',
                ((context,) for context in contexts)
            )
            context_ids = {
                context: id
                for id, context in conn.execute('select id, context from context')
            }

            # Prepare tracers and fail, if a conflict is found.
            # tracer_paths is used to ensure consistency over the tracer data
            # and tracer_map tracks the tracers to be inserted.
            tracer_map = {}
            for path in files.values():
                this_tracer = this_tracers.get(path)
                other_tracer = tracers.get(path, '')
                # If there is no tracer, there is always the None tracer.
                if this_tracer is not None and this_tracer != other_tracer:
                    raise CoverageException(
                        "Conflicting file tracer name for '%s': %r vs %r" % (
                            path, this_tracer, other_tracer
                        )
                    )
                tracer_map[path] = other_tracer

            # Prepare arc and line rows to be inserted by converting the file
            # and context strings with integer ids. Then use the efficient
            # `executemany()` to insert all rows at once.
            arc_rows = (
                (file_ids[file], context_ids[context], fromno, tono)
                for file, context, fromno, tono in arcs
            )

            # Get line data.
            cur = conn.execute(
                'select file.path, context.context, line_bits.numbits '
                'from line_bits '
                'inner join file on file.id = line_bits.file_id '
                'inner join context on context.id = line_bits.context_id'
                )
            for path, context, numbits in cur:
                key = (aliases.map(path), context)
                lines[key] = numbits
            cur.close()

            if arcs:
                self._choose_lines_or_arcs(arcs=True)

                # Write the combined data.
                conn.executemany(
                    'insert or ignore into arc '
                    '(file_id, context_id, fromno, tono) values (?, ?, ?, ?)',
                    arc_rows
                )

            if lines:
                self._choose_lines_or_arcs(lines=True)
                conn.execute("delete from line_bits")
                conn.executemany(
                    "insert into line_bits "
                    "(file_id, context_id, numbits) values (?, ?, ?)",
                    [
                        (file_ids[file], context_ids[context], numbits)
                        for (file, context), numbits in lines.items()
                    ]
                )
            conn.executemany(
                'insert or ignore into tracer (file_id, tracer) values (?, ?)',
                ((file_ids[filename], tracer)
                 for filename, tracer in tracer_map.items())
            )

        # Update all internal cache data.
        self._reset()
        self.read()

    def erase(self, parallel=False):
        """Erase the data in this object.

        If `parallel` is true, then also deletes data files created from the
        basename by parallel-mode.

        """
        self._reset()
        if self._no_disk:
            return
        if self._debug.should('dataio'):
            self._debug.write("Erasing data file {!r}".format(self._filename))
        file_be_gone(self._filename)
        if parallel:
            data_dir, local = os.path.split(self._filename)
            localdot = local + '.*'
            pattern = os.path.join(os.path.abspath(data_dir), localdot)
            for filename in glob.glob(pattern):
                if self._debug.should('dataio'):
                    self._debug.write(
                        "Erasing parallel data file {!r}".format(filename))
                file_be_gone(filename)

    def read(self):
        """Start using an existing data file."""
        with self._connect():       # TODO: doesn't look right
            self._have_used = True

    def write(self):
        """Ensure the data is written to the data file."""

    def _start_using(self):
        """Call this before using the database at all."""
        if self._pid != os.getpid():
            # Looks like we forked! Have to start a new data file.
            self._reset()
            self._choose_filename()
            self._pid = os.getpid()
        if not self._have_used:
            self.erase()
        self._have_used = True

    def has_arcs(self):
        """Does the database have arcs (True) or lines (False)."""
        return bool(self._has_arcs)

    def measured_files(self):
        """A set of all files that had been measured."""
        return set(self._file_map)

    def measured_contexts(self):
        """A set of all contexts that have been measured.

        .. versionadded:: 5.0

        """
        self._start_using()
        with self._connect() as con:
            contexts = {
                row[0]
                for row in con.execute("select distinct(context) from context")
            }
        return contexts

    def file_tracer(self, filename):
        """Get the plugin name of the file tracer for a file.

        Returns the name of the plugin that handles this file.  If the file was
        measured, but didn't use a plugin, then "" is returned.  If the file
        was not measured, then None is returned.

        """
        self._start_using()
        with self._connect() as con:
            file_id = self._file_id(filename)
            if file_id is None:
                return None
            row = con.execute_one(
                "select tracer from tracer where file_id = ?", (file_id,))
            if row is not None:
                return row[0] or ""
            return ""   # File was measured, but no tracer associated.

    def set_query_context(self, context):
        """Set a context for subsequent querying.

        The next :meth:`lines`, :meth:`arcs`, or :meth:`contexts_by_lineno`
        calls will be limited to only one context.  `context` is a string which
        must match a context exactly.  If it does not, no exception is raised,
        but queries will return no data.

        .. versionadded:: 5.0

        """
        self._start_using()
        with self._connect() as con:
            cur = con.execute("select id from context where context = ?", (context,))
            self._query_context_ids = [row[0] for row in cur.fetchall()]

    def set_query_contexts(self, contexts):
        """Set a number of contexts for subsequent querying.

        The next :meth:`lines`, :meth:`arcs`, or :meth:`contexts_by_lineno`
        calls will be limited to the specified contexts.  `contexts` is a list
        of Python regular expressions.  Contexts will be matched using
        :func:`re.search <python:re.search>`.  Data will be included in query
        results if they are part of any of the contexts matched.

        .. versionadded:: 5.0

        """
        self._start_using()
        if contexts:
            with self._connect() as con:
                context_clause = ' or '.join(['context regexp ?'] * len(contexts))
                cur = con.execute(
                    "select id from context where " + context_clause, contexts)
                self._query_context_ids = [row[0] for row in cur.fetchall()]
        else:
            self._query_context_ids = None

    def lines(self, filename):
        """Get the list of lines executed for a file.

        If the file was not measured, returns None.  A file might be measured,
        and have no lines executed, in which case an empty list is returned.

        If the file was executed, returns a list of integers, the line numbers
        executed in the file. The list is in no particular order.

        """
        self._start_using()
        if self.has_arcs():
            arcs = self.arcs(filename)
            if arcs is not None:
                all_lines = itertools.chain.from_iterable(arcs)
                return list({l for l in all_lines if l > 0})

        with self._connect() as con:
            file_id = self._file_id(filename)
            if file_id is None:
                return None
            else:
                query = "select numbits from line_bits where file_id = ?"
                data = [file_id]
                if self._query_context_ids is not None:
                    ids_array = ', '.join('?' * len(self._query_context_ids))
                    query += " and context_id in (" + ids_array + ")"
                    data += self._query_context_ids
                bitmaps = list(con.execute(query, data))
                nums = set()
                for row in bitmaps:
                    nums.update(numbits_to_nums(row[0]))
                return list(nums)

    def arcs(self, filename):
        """Get the list of arcs executed for a file.

        If the file was not measured, returns None.  A file might be measured,
        and have no arcs executed, in which case an empty list is returned.

        If the file was executed, returns a list of 2-tuples of integers. Each
        pair is a starting line number and an ending line number for a
        transition from one line to another. The list is in no particular
        order.

        Negative numbers have special meaning.  If the starting line number is
        -N, it represents an entry to the code object that starts at line N.
        If the ending ling number is -N, it's an exit from the code object that
        starts at line N.

        """
        self._start_using()
        with self._connect() as con:
            file_id = self._file_id(filename)
            if file_id is None:
                return None
            else:
                query = "select distinct fromno, tono from arc where file_id = ?"
                data = [file_id]
                if self._query_context_ids is not None:
                    ids_array = ', '.join('?' * len(self._query_context_ids))
                    query += " and context_id in (" + ids_array + ")"
                    data += self._query_context_ids
                arcs = con.execute(query, data)
                return list(arcs)

    def contexts_by_lineno(self, filename):
        """Get the contexts for each line in a file.

        Returns:
            A dict mapping line numbers to a list of context names.

        .. versionadded:: 5.0

        """
        lineno_contexts_map = collections.defaultdict(list)
        self._start_using()
        with self._connect() as con:
            file_id = self._file_id(filename)
            if file_id is None:
                return lineno_contexts_map
            if self.has_arcs():
                query = (
                    "select arc.fromno, arc.tono, context.context "
                    "from arc, context "
                    "where arc.file_id = ? and arc.context_id = context.id"
                )
                data = [file_id]
                if self._query_context_ids is not None:
                    ids_array = ', '.join('?' * len(self._query_context_ids))
                    query += " and arc.context_id in (" + ids_array + ")"
                    data += self._query_context_ids
                for fromno, tono, context in con.execute(query, data):
                    if context not in lineno_contexts_map[fromno]:
                        lineno_contexts_map[fromno].append(context)
                    if context not in lineno_contexts_map[tono]:
                        lineno_contexts_map[tono].append(context)
            else:
                query = (
                    "select l.numbits, c.context from line_bits l, context c "
                    "where l.context_id = c.id "
                    "and file_id = ?"
                    )
                data = [file_id]
                if self._query_context_ids is not None:
                    ids_array = ', '.join('?' * len(self._query_context_ids))
                    query += " and l.context_id in (" + ids_array + ")"
                    data += self._query_context_ids
                for numbits, context in con.execute(query, data):
                    for lineno in numbits_to_nums(numbits):
                        lineno_contexts_map[lineno].append(context)
        return lineno_contexts_map

    @classmethod
    def sys_info(cls):
        """Our information for `Coverage.sys_info`.

        Returns a list of (key, value) pairs.

        """
        with SqliteDb(":memory:", debug=NoDebugging()) as db:
            temp_store = [row[0] for row in db.execute("pragma temp_store")]
            compile_options = [row[0] for row in db.execute("pragma compile_options")]

        return [
            ('sqlite3_version', sqlite3.version),
            ('sqlite3_sqlite_version', sqlite3.sqlite_version),
            ('sqlite3_temp_store', temp_store),
            ('sqlite3_compile_options', compile_options),
        ]


class SqliteDb(SimpleReprMixin):
    """A simple abstraction over a SQLite database.

    Use as a context manager, then you can use it like a
    :class:`python:sqlite3.Connection` object::

        with SqliteDb(filename, debug_control) as db:
            db.execute("insert into schema (version) values (?)", (SCHEMA_VERSION,))

    """
    def __init__(self, filename, debug):
        self.debug = debug if debug.should('sql') else None
        self.filename = filename
        self.nest = 0
        self.con = None

    def _connect(self):
        """Connect to the db and do universal initialization."""
        if self.con is not None:
            return

        # SQLite on Windows on py2 won't open a file if the filename argument
        # has non-ascii characters in it.  Opening a relative file name avoids
        # a problem if the current directory has non-ascii.
        filename = self.filename

        # It can happen that Python switches threads while the tracer writes
        # data. The second thread will also try to write to the data,
        # effectively causing a nested context. However, given the idempotent
        # nature of the tracer operations, sharing a connection among threads
        # is not a problem.
        if self.debug:
            self.debug.write("Connecting to {!r}".format(self.filename))
        self.con = sqlite3.connect(filename, check_same_thread=False)
        self.con.create_function('REGEXP', 2, _regexp)

        # This pragma makes writing faster. It disables rollbacks, but
        # we never need them.  PyPy needs the .close() calls here, or
        # sqlite gets twisted up:
        # https://bitbucket.org/pypy/pypy/issues/2872/default-
        # isolation-mode-is-different-on
        self.execute("pragma journal_mode=off").close()
        # This pragma makes writing faster.
        self.execute("pragma synchronous=off").close()

    def close(self):
        """If needed, close the connection."""
        if self.con is not None and self.filename != ":memory:":
            self.con.close()
            self.con = None

    def __enter__(self):
        if self.nest == 0:
            self._connect()
            self.con.__enter__()
        self.nest += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.nest -= 1
        if self.nest == 0:
            try:
                self.con.__exit__(exc_type, exc_value, traceback)
                self.close()
            except Exception as exc:
                if self.debug:
                    self.debug.write("EXCEPTION from __exit__: {}".format(exc))
                raise

    def execute(self, sql, parameters=()):
        """Same as :meth:`python:sqlite3.Connection.execute`."""
        if self.debug:
            tail = " with {!r}".format(parameters) if parameters else ""
            self.debug.write("Executing {!r}{}".format(sql, tail))
        try:
            try:
                return self.con.execute(sql, parameters)
            except Exception:
                # In some cases, an error might happen that isn't really an
                # error.  Try again immediately.
                # https://github.com/nedbat/coveragepy/issues/1010
                return self.con.execute(sql, parameters)
        except sqlite3.Error as exc:
            msg = str(exc)
            try:
                # `execute` is the first thing we do with the database, so try
                # hard to provide useful hints if something goes wrong now.
                with open(self.filename, "rb") as bad_file:
                    cov4_sig = b"!coverage.py: This is a private format"
                    if bad_file.read(len(cov4_sig)) == cov4_sig:
                        msg = (
                            "Looks like a coverage 4.x data file. "
                            "Are you mixing versions of coverage?"
                        )
            except Exception:
                pass
            if self.debug:
                self.debug.write("EXCEPTION from execute: {}".format(msg))
            raise CoverageException(
                "Couldn't use data file {!r}: {}".format(self.filename, msg))

    def execute_one(self, sql, parameters=()):
        """Execute a statement and return the one row that results.

        This is like execute(sql, parameters).fetchone(), except it is
        correct in reading the entire result set.  This will raise an
        exception if more than one row results.

        Returns a row, or None if there were no rows.
        """
        rows = list(self.execute(sql, parameters))
        if len(rows) == 0:
            return None
        elif len(rows) == 1:
            return rows[0]
        else:
            raise CoverageException(
                "Sql {!r} shouldn't return {} rows".format(sql, len(rows)))

    def executemany(self, sql, data):
        """Same as :meth:`python:sqlite3.Connection.executemany`."""
        if self.debug:
            data = list(data)
            self.debug.write(
                "Executing many {!r} with {} rows".format(sql, len(data)))
        return self.con.executemany(sql, data)

    def executescript(self, script):
        """Same as :meth:`python:sqlite3.Connection.executescript`."""
        self.con.executescript(script)

    def dump(self):
        """Return a multi-line string, the SQL dump of the database."""
        return "\n".join(self.con.iterdump())


def _regexp(text, pattern):
    """A regexp function for SQLite."""
    return re.search(text, pattern) is not None



from ..parser import ast


class HasYieldVisitor(ast.NodeVisitor):

    def __init__(self):
        self.has_yield = False

    def visit_Yield(self, _node):
        self.has_yield = True


def has_yield(node):
    visitor = HasYieldVisitor()
    visitor.visit(node)

    return visitor.has_yield


def create_assign_state_node(state):
    return ast.Assign(
        targets=[ast.Attribute(value=ast.Name(id='self'), attr='_state')],
        value=ast.Constant(value=state))


class IteratorVisitor(ast.NodeVisitor):

    def __init__(self):
        self._next_state = 0
        self._resume_states_stack = []
        self._current_states_stack = []
        self._state_bodies = {}
        self._loop_states_stack = []

    def new_state(self):
        state = self._next_state
        self._next_state += 1

        return state

    def push_resume_state(self, state):
        self._resume_states_stack.append(state)

    def pop_resume_state(self):
        return self._resume_states_stack.pop()

    def get_resume_state(self):
        return self._resume_states_stack[-1]

    def push_current_state(self, state):
        self._current_states_stack.append(state)

    def pop_current_state(self):
        return self._current_states_stack.pop()

    def get_current_state(self):
        return self._current_states_stack[-1]

    def push_loop_states(self, states):
        self._loop_states_stack.append(states)

    def pop_loop_states(self):
        return self._loop_states_stack.pop()

    def get_loop_begin_state(self):
        return self._loop_states_stack[-1][0]

    def get_loop_end_state(self):
        return self._loop_states_stack[-1][1]

    def append_to_current_state(self, nodes):
        current_state = self.get_current_state()
        body = self._state_bodies.get(current_state)

        if body is None:
            self._state_bodies[current_state] = nodes
        else:
            body += nodes

    def members(self):
        return []

    def init_args(self):
        return []

    def init_body(self):
        return []

    def match_cases(self):
        cases = []

        for state, body in self._state_bodies.items():
            cases.append(ast.match_case(pattern=ast.Constant(value=state),
                                        body=body))

        return cases

    def visit_body(self, body_nodes):
        body_before_yield = []
        body_iter = iter(body_nodes)
        yield_found = False

        for body_node in body_iter:
            if has_yield(body_node):
                self.push_resume_state(self.new_state())
                body_before_yield += self.visit(body_node)
                self.push_current_state(self.get_resume_state())
                self.pop_resume_state()

                for body_node in body_iter:
                    if has_yield(body_node):
                        self.push_resume_state(self.new_state())
                        self.append_to_current_state(self.visit(body_node))
                        self.pop_current_state()
                        self.push_current_state(self.pop_resume_state())
                    else:
                        self.append_to_current_state([body_node])

                self.append_to_current_state([
                    create_assign_state_node(self.get_resume_state())
                ])
                self.pop_current_state()
                yield_found = True
            else:
                body_before_yield.append(body_node)

        if not yield_found:
            body_before_yield.append(
                create_assign_state_node(self.get_resume_state()))

        return body_before_yield

    def visit_If(self, node):
        return [ast.If(test=node.test,
                       body=self.visit_body(node.body),
                       orelse=self.visit_body(node.orelse))]

    def visit_FunctionDef(self, node):
        self.push_current_state(self.new_state())

        for body_node in node.body:
            if has_yield(body_node):
                self.push_resume_state(self.new_state())
                self.append_to_current_state(self.visit(body_node))
                self.pop_current_state()
                self.push_current_state(self.pop_resume_state())
            else:
                self.append_to_current_state([body_node])

        self.append_to_current_state([
            ast.Raise(
                exc=ast.Call(
                    func=ast.Name(id='RuntimeError'),
                    args=[],
                    keywords=[]))
        ])

    def visit_Yield(self, node):
        return [
            create_assign_state_node(self.get_resume_state()),
            ast.Return(value=node.value)
        ]

    def visit_While(self, node):
        while_state = self.new_state()
        self.push_current_state(while_state)
        self.push_loop_states((while_state, self.get_resume_state()))
        self.push_resume_state(while_state)
        body = self.visit_body(node.body)
        self.pop_resume_state()
        self.pop_loop_states()
        self.append_to_current_state([
            ast.If(test=node.test,
                   body=body,
                   orelse=[
                       create_assign_state_node(self.get_resume_state())
                   ])
        ])
        self.pop_current_state()

        return [create_assign_state_node(while_state)]

    def visit_For(self, node):
        raise NotImplementedError("'for' is not yet implemented in iterators")

    def visit_Try(self, node):
        raise NotImplementedError("'try' is not yet implemented in iterators")

    def visit_Expr(self, node):
        return [ast.Expr(value=node) for node in self.visit(node.value)]

    def visit_Return(self, node):
        raise Exception("'return' is not allowed in iterators")


def transform(tree):
    """Transform given iterator function AST tree to an iterator
    class. The class is returned.

    """

    visitor = IteratorVisitor()
    visitor.visit(tree)

    return ast.ClassDef(
        name=tree.name.title(),
        bases=[],
        keywords=[],
        body=[
            ast.AnnAssign(
                target=ast.Name(id='_state'),
                annotation=ast.Name(id='i64'),
                simple=1)
        ] + visitor.members() + [
            ast.FunctionDef(
                name='__init__',
                args=ast.arguments(
                    posonlyargs=[],
                    args=[ast.arg(arg='self')] + tree.args.args,
                    kwonlyargs=tree.args.kwonlyargs,
                    kw_defaults=tree.args.kw_defaults,
                    defaults=tree.args.defaults),
                body=[create_assign_state_node(0)] + visitor.init_body(),
                decorator_list=[]),
            ast.FunctionDef(
                name='next',
                args=ast.arguments(posonlyargs=[],
                                   args=[ast.arg(arg='self')],
                                   kwonlyargs=[],
                                   kw_defaults=[],
                                   defaults=[]),
                body=[
                    ast.While(
                        test=ast.Constant(value=True),
                        body=ast.Match(
                            subject=ast.Attribute(
                                value=ast.Name(id='self'), attr='_state'),
                            cases=visitor.match_cases()),
                        orelse=[])
                ],
                decorator_list=[],
                returns=ast.Name(id='i64'))
        ],
        decorator_list=[])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import toml

# -- Project information -----------------------------------------------------

project = '{package_name}'
# copyright = ''
author = '{authors}'

# The full version, including alpha/beta/rc tags

with open('../package.toml') as fin:
    config = toml.loads(fin.read())

version = config['package']['version']
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx_rtd_theme',
    'mys.sphinx_rtd_dark_mode',
    'mys.sphinx'
]

# user starts in light mode
default_dark_mode = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

html_theme_options = {{
    'style_nav_header_background': '#7289da'
}}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
master_doc = 'index'

# autosummary_generate = True

# html_logo = '_static/logo.png'
# html_favicon = '_static/logo.png'
